# Copyright Commvault Systems, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# --------------------------------------------------------------------------

"""Main file for performing resource pool related operations on CS

ResourcePools , ResourcePool and ResourcePoolTypes are the classes defined in this file

ResourcePools:

        __init__()                          --  initialise object of the ResourcePools class

        _response_not_success()             --  parses through the exception response, and raises SDKException

        _get_resource_pools()               --  returns resource pools details from CS

        has()                               --  Checks whether given resource pool exists in cs or not

        get()                               -- returns ResourcePool object for given name

        delete()                            --  deletes the resource pool from CS

        create()                            --  creates resource pool in CS

        refresh()                           --  Refreshes resource pools associated with cs

ResourcePool:

        __init__()                          --  initialise object of the ResourcePool class

        _response_not_success()             --  parses through the exception response, and raises SDKException

        _get_pool_details()                 --  returns resource pool details from cs

        refresh()                           --  refreshes resource pool details

ResourcePool Attributes:
----------------------------------

    **resource_pool_id**        --  returns Resource pool id

    **resource_pool_type**      --  returns ResourcePoolTypes enum

"""

from .exception import SDKException

import enum


class ResourcePoolTypes(enum.Enum):
    """Enum class for different resource pool types"""
    GENERIC = 0
    O365 = 1
    SALESFORCE = 2
    EXCHANGE = 3
    SHAREPOINT = 4
    ONEDRIVE = 5
    TEAMS = 6
    DYNAMICS_365 = 7
    VSA = 8
    FILESYSTEM = 9
    KUBERNETES = 10
    AZURE_AD = 11
    CLOUD_LAPTOP = 12
    FILE_STORAGE_OPTIMIZATION = 13
    DATA_GOVERNANCE = 14
    E_DISCOVERY = 15
    CLOUD_DB = 16
    OBJECT_STORAGE = 17
    GMAIL = 18
    GOOGLE_DRIVE = 19
    GOOGLE_WORKSPACE = 20
    SERVICENOW = 21
    THREATSCAN = 22
    DEVOPS = 23
    RISK_ANALYSIS = 24
    GOOGLE_CLOUD_PLATFORM = 50001


class ResourcePools():
    """class to represent all pool in resource pool"""

    def __init__(self, commcell_object):
        """Initializes an instance of the ResourcePools class.

            Args:

                commcell_object     (object)    --  instance of the commcell class

            Returns:

                object  -   instance of the ResourcePools class

        """
        self._commcell_object = commcell_object
        self._cvpysdk_object = commcell_object._cvpysdk_object
        self._services = commcell_object._services
        self._API_GET_ALL_RESOURCE_POOLS = self._services['GET_RESOURCE_POOLS']
        self._API_DELETE_RESOURCE_POOL = self._services['DELETE_RESOURCE_POOL']
        self._API_CREATE_RESOURCE_POOL = self._services['CREATE_RESOURCE_POOL']
        self._pools = {}
        self.refresh()

    def _response_not_success(self, response):
        """Helper function to raise an exception when reponse status is not 200 (OK).

            Args:
                response    (object)    --  response class object,

                received upon running an API request, using the `requests` python package

        """
        raise SDKException('Response', '101', self._commcell_object._update_response_(response.text))

    def _get_resource_pools(self) -> dict:
        """returns resource pools details from CS

                Args:

                    None

                Returns:

                    dict       --  Resource pool details

                Raises:

                    SDKException:

                        if failed to get resource pool details

        """
        flag, response = self._cvpysdk_object.make_request('GET', self._API_GET_ALL_RESOURCE_POOLS)
        output = {}
        if flag:
            if response.json() and 'resourcePools' in response.json():
                _resourcepools = response.json()['resourcePools']
                for _pool in _resourcepools:
                    if 'name' in _pool:
                        output.update({_pool['name'].lower(): _pool})
            elif bool(response.json()):
                raise SDKException('ResourcePools', '103')
            return output
        self._response_not_success(response)

    def create(self, name: str, resource_type, **kwargs):
        """creates resource pool in CS

            Args:

                name    (str)           --  Resource pool name

                resource_type   (enum)  --  ResourcePoolTypes enum

            kwargs options:

                for Threat Scan -

                    index_server    (str)   --  index server name

            Returns:

                obj --  Instance of ResourcePool class

            Raises:

                SDKException:

                    if failed to create resource pool

                    if resource pool already exists

        """
        if resource_type.value not in [ResourcePoolTypes.THREATSCAN.value]:
            raise SDKException('ResourcePools', '102', 'Resource pool creation is not supported for this resource type')
        if resource_type.value == ResourcePoolTypes.THREATSCAN.value and 'index_server' not in kwargs:
            raise SDKException('ResourcePools', '102', 'Index server name is missing in kwargs')
        if self.has(name=name):
            raise SDKException('ResourcePools', '107')
        _request_json = {
            "resourcePool": {
                "appType": resource_type.value,
                "dataAccessNodes": [],
                "extendedProp": {
                    "exchangeOnePassClientProperties": {}},
                "resourcePool": {
                    "resourcePoolId": 0,
                    "resourcePoolName": name},
                "exchangeServerProps": {
                    "jobResultsDirCredentials": {
                        "userName": ""},
                    "jobResultsDirPath": ""},
                "roleId": None,
                "indexServerMembers": [],
                "indexServer": {
                    "clientId": self._commcell_object.index_servers.get(
                        kwargs.get('index_server')).index_server_client_id if resource_type.value == ResourcePoolTypes.THREATSCAN.value else 0,
                    "clientName": kwargs.get('index_server') if resource_type.value == ResourcePoolTypes.THREATSCAN.value else '',
                    "displayName": kwargs.get('index_server') if resource_type.value == ResourcePoolTypes.THREATSCAN.value else '',
                    "selected": True},
                "accessNodes": {
                    "clientGroups": [],
                    "clients": []}}}
        flag, response = self._cvpysdk_object.make_request(
            'POST', self._API_CREATE_RESOURCE_POOL, _request_json)
        if flag:
            if response.json() and 'error' in response.json():
                _error = response.json()['error']
                if _error.get('errorCode', 0) != 0:
                    raise SDKException('ResourcePools', '102', f'Resource pool creation failed with {_error}')
                self.refresh()
                return self.get(name=name)
            raise SDKException('ResourcePools', '108')
        self._response_not_success(response)

    def delete(self, name: str):
        """deletes the resource pool from CS

            Args:

                name   (str)       --   Resource pool name

            Returns:

                None

            Raises:

                SDKException:

                    if failed to delete resource pool

                    if failed to find resource pool
        """
        if not self.has(name=name):
            raise SDKException('ResourcePools', '104')
        api = self._API_DELETE_RESOURCE_POOL % (self._pools[name.lower()]['id'])
        flag, response = self._cvpysdk_object.make_request('DELETE', api)
        if flag:
            if response.json() and 'error' in response.json():
                _error = response.json()['error']
                if _error.get('errorCode', 0) != 0:
                    raise SDKException('ResourcePools', '102', f'Resource pool deletion failed with {_error}')
                self.refresh()
                return
            raise SDKException('ResourcePools', '106')
        self._response_not_success(response)

    def get(self, name: str):
        """returns ResourcePool object for given name

            Args:

                Name    (str)       -- Resource Pool name

            Returns:

                obj     -- Instance of ResourcePool class

            Raises:

                SDKException:

                    if failed to find resource pool with given name

        """
        if not self.has(name):
            raise SDKException('ResourcePools', '104')
        return ResourcePool(commcell_object=self._commcell_object, name=name, pool_id=self._pools[name.lower()]['id'])

    def has(self, name: str) -> bool:
        """Checks whether given resource pool exists in cs or not

            Args:

                name        (str)       -- Resource pool name

            Returns:

               bool    --  True if resource pool exists in cs
        """
        if name.lower() in self._pools:
            return True
        return False

    def refresh(self):
        """Refresh the resource pools associated with CS"""
        self._pools = {}
        self._pools = self._get_resource_pools()


class ResourcePool:

    def __init__(self, commcell_object, name, pool_id):
        """Initializes an instance of the ResourcePool class.

            Args:

                commcell_object     (object)    --  instance of the commcell class

                server_name         (str)       --  Name of the resurce pool

                pool_id             (str)       --  Resource pool id


            Returns:

                object  -   instance of the ResourcePool class

        """
        self._commcell_object = commcell_object
        self._update_response_ = commcell_object._update_response_
        self._cvpysdk_object = commcell_object._cvpysdk_object
        self._services = commcell_object._services
        self._resource_pool_name = name
        self._resource_pool_id = pool_id
        self._resource_details = None
        self._API_GET_POOL_DETAILS = self._services['GET_RESOURCE_POOL_DETAILS']
        self.refresh()

    def _response_not_success(self, response):
        """Helper function to raise an exception when reponse status is not 200 (OK).

            Args:
                response    (object)    --  response class object,

                received upon running an API request, using the `requests` python package

        """
        raise SDKException('Response', '101', self._commcell_object._update_response_(response.text))

    def _get_pool_details(self) -> dict:
        """returns resource pool details from CS

            Args:

                None

            Returns:

                dict        - Resource pool details

            Raises:

                SDKException:

                    if failed to get details for resource pool
        """
        api = self._API_GET_POOL_DETAILS % (self._resource_pool_id)
        flag, response = self._cvpysdk_object.make_request('GET', api)
        if flag:
            if response.json() and 'resourcePool' in response.json():
                return response.json()['resourcePool']
            raise SDKException('ResourcePools', '105')
        self._response_not_success(response)

    def refresh(self):
        """Refresh the resource pool details"""
        self._resource_details = None
        self._resource_details = self._get_pool_details()

    @property
    def resource_pool_id(self):
        """returns the pool id for this resource pool

            Returns:

                int --  resource pool id

        """
        return int(self._resource_details['resourcePool'].get('resourcePoolId'))

    @property
    def resource_pool_type(self):
        """returns the pool type enum for this resource pool

            Returns:

                enum --  ResourcePoolTypes

        """
        return ResourcePoolTypes(int(self._resource_details['appType']))



# -*- coding: utf-8 -*-

# --------------------------------------------------------------------------
# Copyright Commvault Systems, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# --------------------------------------------------------------------------

"""File for operating on a Virtual Server Azure Resource Manager Instance.

AzureResoureceManagerInstance is the only class defined in this file.

AzureResoureceManagerInstance: Derived class from VirtualServer
                            Base class, representing a Azure Resource Manager
                            instance, and to perform operations on that
                            instance

    __init__(self, agent,_name,iid)   	 -- 	initialize object of azure RM
                                            Instance object associated with the
                                            VirtualServer Instance		

    _get_instance_properties()     --  VirtualServer Instance class method
                                        overwritten to get azure RM
                                        Specific instance properties as well

	_get_instance_properties_json()			--  get the all instance related
														properties of this subclient.

"""

from ..vsinstance import VirtualServerInstance
from ...exception import SDKException
from ...instance import Instance

class AzureRMInstance(VirtualServerInstance):

    '''
    AzureResoureceManagerInstance:

	__init__(agent_object,instance_name,instance_id)    --  initialize object
    of azure resource manager Instance object associated with
		the VirtualServer Instance
    '''

    def __init__(self, agent, name, iid):
        """Initialize the Instance object for the given Virtual Server instance.

            Args:
                class_object (agent_object,instance_name,instance_id)  --  instance of the
                                                                                Agent class,
                                                                                instance name,
                                                                                instance id

        """


        super(VirtualServerInstance, self).__init__(agent, name, iid)
        self._vendor_id = 7
        self._subscription_id = None



    def  _get_instance_properties(self):
        """
        Get the properties of this instance

        Raise:
            SDK Exception:
                if response is not empty
                if response is not success
        """

        super(AzureRMInstance, self)._get_instance_properties()
        self._server_name = []
        if 'virtualServerInstance' in self._properties:
            if self._properties["virtualServerInstance"]["associatedClients"].get("memberServers"):
                _member_servers = self._properties["virtualServerInstance"] \
                                                ["associatedClients"]["memberServers"]
            else:
                _member_servers=[]
            for _each_client in _member_servers:
                client = _each_client['client']
                if 'clientName' in client.keys():
                    self._server_name.append(str(client['clientName']))
        # waiting for praveen form


    def _get_instance_properties_json(self):
        """get the all instance related properties of this subclient.

          Returns:
               dict - all instance properties put inside a dict

        """
        instance_json = {
            "instanceProperties":{
                "isDeleted": False,
                "instance": self._instance,
                "instanceActivityControl": self._instanceActivityControl,
                "virtualServerInstance": {
                    "vsInstanceType": self._virtualserverinstance['vsInstanceType'],
                    "associatedClients": self._virtualserverinstance['associatedClients'],
                    "vmwareVendor": {}
                    }
                       }
               }
        return instance_json

    def _get_application_properties(self):
        """
            Get the properties of this instance

            Raise:
                SDK Exception:
                    if response is not empty
                    if response is not success
        """
        super(AzureRMInstance, self)._get_application_properties()
        if 'azureResourceManager' in self._application_properties:
            self._subscription_id = self._application_properties['azureResourceManager']['subscriptionId']

    def _update_azure_credentials(self, credential_id, credential_name=None, usemanaged_identity=False):
        """
        To update the credentials in azure hypervisor
        Args:
                credential_id (int)  --  Credential ID to update in hypervisor
                credential_name(str) -- Credential name to update in hypervisor
                usemanaged_identity( bool) -- to use managed identity
        """

        self._get_application_properties()

        self._credential_json = {
            "hypervisorType": self._vendor_id,
            "skipCredentialValidation": False,
            "credentials": {
                "id": credential_id,
                "name": credential_name
            },
            "subscriptionId": self._subscription_id,
         "useManagedIdentity": usemanaged_identity
        }

        super(AzureRMInstance, self)._update_hypervisor_credentials(self._credential_json)


    @property
    def server_name(self):
        """getter for the domain name in the Hyper-V json"""
        return self._server_name

    @property
    def server_host_name(self):
        """getter for the domain name in the vmware vendor json"""
        return self._server_name
    #return self._server_name
    # TODO will change with Praveen Form(jmalik)



# -*- coding: utf-8 -*-

# --------------------------------------------------------------------------
# Copyright Commvault Systems, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# --------------------------------------------------------------------------

"""File for operating on a Google Instance.

GoogleInstance is the only class defined in this file.

GoogleInstance: Derived class from CloudAppsInstance Base class, representing a
Google (GMail/GDrive) and OneDrive instance,
and to perform operations on that instance

GoogleInstance:

    _prepare_restore_json_v2()  --  Utility function to prepare user level restore json for
                                    OneDrive for bussiness clients

    _get_instance_properties()  --  Instance class method overwritten to add cloud apps
    instance properties as well

    restore_out_of_place()      --  runs out-of-place restore for the instance

    modify_index_server()       --  Method to modify the index server

    modify_accessnodes()        --  Method to modify accessnodes

"""

from __future__ import unicode_literals
from ...exception import SDKException
from ..cainstance import CloudAppsInstance
from ...constants import AppIDAType
from base64 import b64encode


class GoogleInstance(CloudAppsInstance):
    """Class for representing an Instance of the GMail/Gdrive instance type."""

    def _get_instance_properties(self):
        """Gets the properties of this instance.

            Raises:
                SDKException:
                    if response is empty

                    if response is not success

        """
        super(GoogleInstance, self)._get_instance_properties()
        # Common properties for Google and OneDrive
        self._ca_instance_type = None
        self._manage_content_automatically = None
        self._auto_discovery_enabled = None
        self._auto_discovery_mode = None
        self._proxy_client = None

        # Google instance related properties
        self._app_email_id = None
        self._google_admin_id = None
        self._service_account_key_file = None
        self._app_client_id = None

        # OneDrive instance related properties
        self._client_id = None
        self._tenant = None

        if 'cloudAppsInstance' in self._properties:
            cloud_apps_instance = self._properties['cloudAppsInstance']
            self._ca_instance_type = cloud_apps_instance['instanceType']

            if 'gInstance' in cloud_apps_instance:
                ginstance = cloud_apps_instance['gInstance']

                self._manage_content_automatically = ginstance['manageContentAutomatically']
                self._auto_discovery_enabled = ginstance['isAutoDiscoveryEnabled']
                self._auto_discovery_mode = ginstance['autoDiscoveryMode']
                self._app_email_id = ginstance['appEmailId']
                self._google_admin_id = ginstance['emailId']
                self._service_account_key_file = ginstance['appKey']
                self._app_client_id = ginstance['appClientId']

            if 'oneDriveInstance' in cloud_apps_instance:
                onedrive_instance = cloud_apps_instance['oneDriveInstance']

                self._manage_content_automatically = onedrive_instance['manageContentAutomatically']
                self._auto_discovery_enabled = onedrive_instance['isAutoDiscoveryEnabled']
                self._auto_discovery_mode = onedrive_instance['autoDiscoveryMode']
                if 'clientId' in onedrive_instance:
                    self._client_id = onedrive_instance.get('clientId')
                    self._tenant = onedrive_instance.get('tenant')
                else:
                    self._client_id = onedrive_instance.get(
                        'azureAppList', {}).get('azureApps', [{}])[0].get('azureAppId')
                    self._tenant = onedrive_instance.get(
                        'azureAppList', {}).get('azureApps', [{}])[0].get('azureDirectoryId')

                if self._client_id is None:
                    raise SDKException('Instance', '102', 'Azure App has not been configured')

            if 'generalCloudProperties' in cloud_apps_instance:
                if 'proxyServers' in cloud_apps_instance['generalCloudProperties']:
                    self._proxy_client = cloud_apps_instance.get(
                        'generalCloudProperties', {}).get('proxyServers', [{}])[0].get('clientName')
                else:
                    if 'clientName' in cloud_apps_instance.get(
                            'generalCloudProperties', {}).get('memberServers', [{}])[0].get('client'):
                        self._proxy_client = cloud_apps_instance.get('generalCloudProperties', {}).get(
                            'memberServers', [{}])[0].get('client', {}).get('clientName')
                    else:
                        self._proxy_client = cloud_apps_instance.get('generalCloudProperties', {}).get(
                            'memberServers', [{}])[0].get('client', {}).get('clientGroupName')

                if self._proxy_client is None:
                    raise SDKException('Instance', '102', 'Access Node has not been configured')

    @property
    def ca_instance_type(self):
        """Returns the CloudApps instance type"""
        if self._ca_instance_type == 1:
            return 'GMAIL'
        elif self._ca_instance_type == 2:
            return 'GDRIVE'
        elif self._ca_instance_type == 7:
            return 'ONEDRIVE'
        return self._ca_instance_type

    @property
    def manage_content_automatically(self):
        """Returns the CloudApps Manage Content Automatically property"""
        return self._manage_content_automatically

    @property
    def auto_discovery_status(self):
        """Treats the Auto discovery property as a read-only attribute."""
        return self._auto_discovery_enabled

    @property
    def auto_discovery_mode(self):
        """Returns the Auto discovery mode property"""
        return self._auto_discovery_mode

    @property
    def app_email_id(self):
        """Returns the service account mail id"""
        return self._app_email_id

    @property
    def google_admin_id(self):
        """Returns the Google admin mail id"""
        return self._google_admin_id

    @property
    def key_file_path(self):
        """Returns the service account key file path"""
        return self._service_account_key_file

    @property
    def google_client_id(self):
        """Returns the service account client id"""
        return self._app_client_id

    @property
    def onedrive_client_id(self):
        """Returns the OneDrive app client id"""
        return self._client_id

    @property
    def onedrive_tenant(self):
        """Returns the OneDrive tenant id"""
        return self._tenant

    @property
    def proxy_client(self):
        """Returns the proxy client name to this instance"""
        return self._proxy_client

    def _prepare_advsearchgrp(self, source_item_list, subclient_id):
        """
                    Utility function to prepare advsearchgrp json for restore job for OneDrive for business clients

                    Args:
                        source_item_list (list)         --  list of user GUID to process in restore

                        subclient_id                    --  subclient id of the client

                    Returns:
                        advsearchgrp (dict) - advsearchgrp json for restore job
        """
        advsearchgrp = {
            "fileFilter": [
                {
                    "interGroupOP": "FTAnd",
                    "filter": {
                        "filters": [
                            {
                                "field": "HIDDEN",
                                "fieldValues": {
                                    "values": [
                                        "true"
                                    ]
                                },
                                "intraFieldOp": "FTNot"
                            },
                            {
                                "field": "CV_OBJECT_GUID",
                                "fieldValues": {
                                    "values": source_item_list
                                },
                                "intraFieldOp": "FTOr"
                            }
                        ],
                        "interFilterOP": "FTAnd"
                    }
                }
            ],
            "commonFilter": [
                {
                    "filter": {
                        "filters": [
                            {
                                "field": "CISTATE",
                                "fieldValues": {
                                    "values": [
                                        "1"
                                    ]
                                },
                                "intraFieldOp": "FTOr",
                                "groupType": 0
                            },
                            {
                                "field": "IS_VISIBLE",
                                "fieldValues": {
                                    "values": [
                                        "true"
                                    ],
                                    "isRange": False,
                                    "isMoniker": False
                                },
                                "intraFieldOp": "FTOr",
                                "intraFieldOpStr": "None"
                            }
                        ],
                        "interFilterOP": "FTAnd"
                    }
                }
            ],
            "galaxyFilter": [
                {
                    "appIdList": [
                        subclient_id
                    ]
                }
            ],
            "graphFilter": [
                {
                    "fromField": "PARENT_GUID",
                    "toField": "CV_OBJECT_GUID",
                    "returnRoot": True,
                    "traversalFilter": [
                        {
                            "filters": [
                                {
                                    "field": "IS_VISIBLE",
                                    "fieldValues": {
                                        "values": [
                                            "true"
                                        ]
                                    },
                                    "intraFieldOp": "FTAnd",
                                    "groupType": 0
                                },
                                {
                                    "field": "HIDDEN",
                                    "fieldValues": {
                                        "values": [
                                            "true"
                                        ]
                                    },
                                    "intraFieldOp": "FTNot"
                                }
                            ]
                        }
                    ]
                }
            ]
        }

        return advsearchgrp

    def _prepare_findquery(self, source_item_list, subclient_id):
        """
            Utility function to prepare findquery json for restore job for OneDrive for bussiness clients

            Args:
                source_item_list (list)         --  list of user GUID to process in restore

                subclient_id                    --  subclient id of the client

            Returns:
                findquery (dict) - findquery json for restore job
        """

        findquery = {
                  "searchProcessingInfo": {
                    "pageSize": 15,
                    "resultOffset": 0,
                    "sortParams": [
                      {
                        "sortField": "DATA_TYPE",
                        "sortDirection": "DESCENDING"
                      },
                      {
                        "sortField": "FileName",
                        "sortDirection": "ASCENDING"
                      }
                    ],
                    "queryParams": [
                      {
                        "param": "ENABLE_MIXEDVIEW",
                        "value": "true"
                      },
                      {
                        "param": "RESPONSE_FIELD_LIST",
                        "value": "FAST_URL,BACKUPTIME,SIZEINKB,MODIFIEDTIME,CONTENTID,CV_TURBO_GUID,AFILEID,AFILEOFFSET,COMMCELLNO,FILE_NAME,FILE_FOLDER,CVSTUB,DATA_TYPE,APPID,JOBID,CISTATE,DATE_DELETED,IdxFlags,CV_OBJECT_GUID,PARENT_GUID,CUSTODIAN,OWNER,ObjectType"
                      },
                      {
                        "param": "DO_NOT_AUDIT",
                        "value": "false"
                      },
                      {
                        "param": "COLLAPSE_FIELD",
                        "value": "CV_OBJECT_GUID"
                      },
                      {
                        "param": "COLLAPSE_SORT",
                        "value": "BACKUPTIME DESC"
                      }
                    ]
                  },
                  "advSearchGrp": self._prepare_advsearchgrp(source_item_list,subclient_id),
                  "mode": "WebConsole"
                }

        return findquery

    def _prepare_restore_json_v2(self, source_item_list, **kwargs):

        """ Utility function to prepare user level restore json for OneDrive for bussiness clients

            Args:
                source_item_list (list)         --  list of user GUID to process in restore

            Kwargs:

                out_of_place (bool)             --  If True, out of place restore will be performed

                disk_restore (bool)             --  If True, restore to disk will be performed

                destination_path (str)          --  destination path for oop and disk restores

                destination_client              -- destination client for disk restore

                overwrite (bool)                --  If True, files will be overwritten in destination if already exists

                restore_as_copy (bool)          --  If True, files will be restored as copy if already exists

                skip_file_permissions (bool)    --  If True, file permissions will be restored

            Returns:
                request_json (dict) - request json for restore job

            Raises:
                SDKException:
                    if destination client with given name does not exist

                    if type of parameter is invalid

        """

        out_of_place = kwargs.get('out_of_place', False)
        disk_restore = kwargs.get('disk_restore', False)
        destination_path = kwargs.get('destination_path', False)
        destination_client = kwargs.get('destination_client')
        overwrite = kwargs.get('overwrite', False)
        restore_as_copy = kwargs.get('restore_as_copy', False)
        skip_file_permissions = kwargs.get('skip_file_permissions', False)



        if destination_client:
            if self._commcell_object.clients.all_clients.get(destination_client):
                destination_client_object = self._commcell_object.clients.all_clients.get(destination_client)
                destination_client_id = int(destination_client_object.get('id'))
            else:
                raise SDKException('Client', '102', 'Client "{0}" does not exist.'.format(destination_client))

        if ((destination_client and not isinstance(destination_client, str) or
             destination_path and not isinstance(destination_path, str)) or not
            (isinstance(source_item_list, list) and
             isinstance(skip_file_permissions, bool) and
             isinstance(disk_restore, bool) and
             isinstance(out_of_place, bool) and
             isinstance(overwrite, bool) and
             isinstance(restore_as_copy, bool))):
            raise SDKException('Instance', '101')

        request_json = self._restore_json(client=self._agent_object._client_object)

        subtasks = request_json['taskInfo']['subTasks'][0]
        options = subtasks['options']
        restore_options = options['restoreOptions']

        restore_options["browseOption"] = {
            "commCellId": self._commcell_object.commcell_id,
            "showDeletedItems": False
        }

        restore_options['commonOptions'] = {
            "overwriteFiles": False,
            "skip": True,
            "unconditionalOverwrite": False
        }

        destination = restore_options['destination']
        destination['destAppId'] = AppIDAType.WINDOWS_FILE_SYSTEM.value if disk_restore else AppIDAType.CLOUD_APP.value
        destination['inPlace'] = False if out_of_place or disk_restore else True

        destination['destClient'] = {
            "clientId": destination_client_id,
            "clientName": destination_client
        } if disk_restore else {
            "clientId": int(self._agent_object._client_object.client_id),
            "clientName": self._agent_object._client_object.client_name
        }

        if destination_path:
            destination['destPath'] = [destination_path]

        restore_options['fileOption']['sourceItem'] = source_item_list

        restore_options['cloudAppsRestoreOptions'] = {
            "instanceType": self._ca_instance_type,
            "googleRestoreOptions": {
                "skipPermissionsRestore": False if disk_restore else skip_file_permissions,
                "restoreToDifferentAccount": True if out_of_place else False,
                "restoreAsCopy": False if disk_restore else restore_as_copy,
                "filelevelRestore": False,
                "strDestUserAccount": destination_path if out_of_place else '',
                "overWriteItems": False if disk_restore else overwrite,
                "restoreToGoogle": False if disk_restore else True
            }
        }

        del subtasks['subTaskOperation']
        del restore_options['fileOption']
        del restore_options['impersonation']
        del restore_options['volumeRstOption']
        del restore_options['sharePointRstOption']
        del restore_options['virtualServerRstOption']

        associations = request_json['taskInfo']['associations'][0]
        subclient_id = associations['subclientId']

        cloudAppsRestoreOptions = restore_options['cloudAppsRestoreOptions']
        cloudAppsRestoreOptions['googleRestoreOptions']['findQuery'] = self._prepare_findquery(source_item_list, subclient_id)

        destination_option = "Destination"
        destination_value = "Original location"
        if out_of_place:
            destination_option = "Destination user"
            destination_value = source_item_list[0]
        if disk_restore:
            destination_option = "Destination server"
            destination_value = destination_client


        options["commonOpts"] = {
            "notifyUserOnJobCompletion": False,
            "jobMetadata": [
              {
                "selectedItems": [
                  {
                    "itemName": source_item_list[0],
                    "itemType": "User"
                  }
                ],
                "jobOptionItems": [
                  {
                    "option": "Restore destination",
                    "value": "OneDrive for Business"
                  },
                  {
                    "option": "Source",
                    "value": source_item_list[0]
                  },
                  {
                    "option": destination_option,
                    "value": destination_value
                  },
                  {
                    "option": "If the file exists",
                    "value": "Skip"
                  },
                  {
                    "option": "Skip file permissions",
                    "value": "Enabled"
                  },
                  {
                    "option": "Include deleted items",
                    "value": "Disabled"
                  }
                ]
              }
            ]
          }

        joboptionitems = options['commonOpts']['jobMetadata'][0]['jobOptionItems']

        if out_of_place:
            joboptionitems.append({"option": "Destination client","value": destination_client })
        if disk_restore:
            joboptionitems.append({"option": "Destination path", "value": destination_path})

        return request_json

    def restore_out_of_place(
            self,
            client,
            destination_path,
            paths,
            overwrite=True,
            restore_data_and_acl=True,
            copy_precedence=None,
            from_time=None,
            to_time=None,
            to_disk=False):
        """Restores the files/folders specified in the input paths list to the input client,
            at the specified destionation location.

            Args:
                client                (str/object) --  either the name of the client or
                                                           the instance of the Client

                destination_path      (str)        --  full path of the restore location on client

                paths                 (list)       --  list of full paths of
                                                           files/folders to restore

                overwrite             (bool)       --  unconditional overwrite files during restore
                    default: True

                restore_data_and_acl  (bool)       --  restore data and ACL files
                    default: True

                copy_precedence         (int)   --  copy precedence value of storage policy copy
                    default: None

                from_time           (str)       --  time to retore the contents after
                        format: YYYY-MM-DD HH:MM:SS

                    default: None

                to_time           (str)         --  time to retore the contents before
                        format: YYYY-MM-DD HH:MM:SS

                    default: None

                to_disk             (bool)       --  If True, restore to disk will be performed

            Returns:
                object - instance of the Job class for this restore job

            Raises:
                SDKException:
                    if client is not a string or Client instance

                    if destination_path is not a string

                    if paths is not a list

                    if failed to initialize job

                    if response is empty

                    if response is not success
        """
        from cvpysdk.client import Client

        if not ((isinstance(client, str) or isinstance(client, Client)) and
                isinstance(destination_path, str) and
                isinstance(paths, list) and
                isinstance(overwrite, bool) and
                isinstance(restore_data_and_acl, bool)):
            raise SDKException('Subclient', '101')

        if isinstance(client, Client):
            client = client
        elif isinstance(client, str):
            client = Client(self._commcell_object, client)
        else:
            raise SDKException('Subclient', '105')

        paths = self._filter_paths(paths)

        destination_path = self._filter_paths([destination_path], True)

        if paths == []:
            raise SDKException('Subclient', '104')

        request_json = self._restore_json(
            paths=paths,
            in_place=False,
            client=client,
            destination_path=destination_path,
            overwrite=overwrite,
            restore_data_and_acl=restore_data_and_acl,
            copy_precedence=copy_precedence,
            from_time=from_time,
            to_time=to_time,
        )
        dest_user_account = destination_path
        rest_different_account = True
        restore_to_google = True

        if to_disk:
            dest_user_account = ''
            rest_different_account = False
            restore_to_google = False
        request_json["taskInfo"]["subTasks"][0]["options"][
            "restoreOptions"]['cloudAppsRestoreOptions'] = {
            "instanceType": self._ca_instance_type,
            "googleRestoreOptions": {
                "strDestUserAccount": dest_user_account,
                "folderGuid": "",
                "restoreToDifferentAccount": rest_different_account,
                "restoreToGoogle": restore_to_google
            }
        }
        return self._process_restore_response(request_json)

    def enable_auto_discovery(self, mode='REGEX'):
        """Enables auto discovery on instance.

           Args:

                mode    (str)   -- Auto Discovery mode

                Valid Values:

                    REGEX
                    GROUP

        """
        auto_discovery_dict = {
            'REGEX': 0,
            'GROUP': 1
        }
        instance_dict = {
            1: 'gInstance',
            2: 'gInstance',
            7: 'oneDriveInstance'
        }
        auto_discovery_mode = auto_discovery_dict.get(mode.upper(), None)

        if auto_discovery_mode is None:
            raise SDKException('Instance', '107')
        instance_prop = self._properties['cloudAppsInstance'].copy()

        instance_prop[instance_dict[instance_prop['instanceType']]]['isAutoDiscoveryEnabled'] = True
        instance_prop[instance_dict[instance_prop['instanceType']]]['autoDiscoveryMode'] = auto_discovery_mode

        self._set_instance_properties("_properties['cloudAppsInstance']", instance_prop)
        self.refresh()

    def _get_instance_properties_json(self):
        """Returns the instance properties json."""

        return {'instanceProperties': self._properties}

    def modify_index_server(self, modified_index_server):
        """
            Method to modify the index server

            Arguments:
                modified_index_server        (str)--     new index server name
        """
        update_dict = {
            "instance": {
                "instanceId": int(self.instance_id),
                "clientId": int(self._agent_object._client_object.client_id),
                "applicationId": int(self._agent_object.agent_id)
            },
                "cloudAppsInstance": {
                    "instanceType": self.ca_instance_type,
                    "oneDriveInstance": {
                    },
                    "generalCloudProperties": {
                        "indexServer": {
                            "clientName": modified_index_server
                        }
                    }
                }
            }

        self.update_properties(properties_dict=update_dict)

    def modify_accessnodes(self,modified_accessnodes_list,modified_user_name,modified_user_password):
        """
                   Method to modify accessnodes

                   Arguments:
                       modified_accessnodes_list     (list)  --     list of new accessnodes
                       modified_user_name            (str)   --     new user account name
                       modified_user_password        (str)   --     new user account password
        """
        member_servers=[]
        for client in modified_accessnodes_list:
            client_dict = {
                "client": {
                    "clientName": client
                }
            }
            member_servers.append(client_dict)

        update_dict = {
            "instance": {
                "instanceId": int(self.instance_id),
                "clientId": int(self._agent_object._client_object.client_id),
                "applicationId": int(self._agent_object.agent_id)
            },
            "cloudAppsInstance": {
                "instanceType": self.ca_instance_type,
                "oneDriveInstance": {
                    "serviceAccounts": {
                        "accounts": [
                            {
                                "serviceType": "SYSTEM_ACCOUNT",
                                "userAccount": {
                                    "userName": modified_user_name,
                                    "password": b64encode(modified_user_password.encode()).decode(),
                                }
                            }
                        ]
                    }
                },
                "generalCloudProperties": {
                    "memberServers": member_servers
                }
            }
        }

        self.update_properties(properties_dict=update_dict)


# -*- coding: utf-8 -*-
# --------------------------------------------------------------------------
# Copyright Commvault Systems, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# --------------------------------------------------------------------------

"""Class to perform all the CommCell Migration operations on commcell

CommCellMigration, GlobalRepositoryCell are the only classes defined in this file.

CommCellMigration: Helper class to perform CommCell Import & Export operations.

CommCellMigration:

    __init__()                      --  initializes CommCellMigration helper object.

    commcell_export()               --  function to run CCM Export operation.

    commcell_import()               --  function to run CCM Import operation.

    tape_import()                   --  function to run tape import operation.

GlobalRepositoryCell: Helper class to perform GRC related operations

GlobalRepositoryCell:

    __init__()                      --  initializes GlobalRepositoryCell object

    get_podcell_entities()          --  gets all entities from registered podcell that can be imported

    get_podcell_properties()        --  gets all grc related properties for registered podcell

    modify_monitored_clients()      --  overwrites imported clients in podcell grc schedule

"""
import html
import xml.etree.ElementTree as ET
from base64 import b64encode
from .job import Job
from .client import Client
from .exception import SDKException


class CommCellMigration(object):
    """Class for representing the commcell export & import operations from commcell. """

    def __init__(self, commcell_object):
        """Initializes object of the CommCellMigration class.

            Args:
               commcell_object (object) -instance of the commcell class

            Returns:
               object - instance of the CommCellMigration class
        """

        self._commcell_object = commcell_object
        self._cvpysdk_object = self._commcell_object._cvpysdk_object
        self._services = self._commcell_object._services
        self._update_response_ = self._commcell_object._update_response_
        self._commcell_name = self._commcell_object.commserv_name
        self._path_type = 0

    def commcell_export(self, export_location, client_list=None, options_dictionary=None, other_entities=None):
        """ Starts the Commcell Export job.

            Args:
                export_location     ( str )         --  Location to export generated dumps.

                client_list         ( list )        --  Contains list of clients used for export.
                    [
                        "Server_1","Client1","Client2"
                    ]

                options_dictionary  ( dict )        --  Contains options used to perform CCM Export.
                    {
                        "pathType":"Local",

                        "otherSqlInstance":True,

                        "userName":"UserName",

                        "password":"User#####",

                        "otherSqlInstance": False,

                        "sqlInstanceName":"SQLInstanceName",

                        "sqlUserName":"SQLUserName",

                        "sqlPassword":"SQLPassword",

                        "Database":"commserv",

                        "captureMediaAgents":True,
                        
                        "captureSchedules":True,

                        "captureActivityControl":True,

                        "captureOperationWindow":True,

                        "captureHolidays":True,

                        "csName": "CommservName",  # host cs for using sql instance export

                        "clientIds": [client_id1, client_id2],  # required only when exporting clients using sql instance

                        "autopickCluster":False
                    }
                
                other_entities      ( list )        --  list of other entities to be exporteddd
                    [
                        "schedule_policies",

                        "users_and_user_groups",

                        "alerts"
                    ]   

            Returns:
                CCM Export Job instance             --  returns the CCM Export job instance.

            Raises:
                SDKException:
                    if type of the input is not valid.

                    if all the required inputs are not provided.

                    if invalid inputs are passed.
        """

        if client_list is None and other_entities is None:
            raise SDKException('CommCellMigration', '105')

        options_dictionary = options_dictionary or {}

        path_type = options_dictionary.get("pathType", "Local")
        network_user_name = options_dictionary.get("userName", "")
        network_user_password = options_dictionary.get("password", "")
        other_sql_instance = options_dictionary.get("otherSqlInstance", False)
        sql_instance_name = options_dictionary.get("sqlInstanceName", "")
        sql_user_name = options_dictionary.get("sqlUserName", "")
        sql_password = options_dictionary.get("sqlPassword", "")
        database = options_dictionary.get("Database", "Commserv")
        capture_ma = options_dictionary.get("captureMediaAgents", True)
        capture_schedules = options_dictionary.get("captureSchedules", True)
        capture_activity_control = options_dictionary.get("captureActivityControl", True)
        capture_opw = options_dictionary.get("captureOperationWindow", True)
        capture_holidays = options_dictionary.get("captureHolidays", True)
        auto_pick_cluster = options_dictionary.get("autopickCluster", False)
        cs_name = options_dictionary.get("csName", self._commcell_name)
        client_ids = options_dictionary.get("clientIds", [])

        if not (isinstance(path_type, str)
                and isinstance(network_user_name, str)
                and isinstance(network_user_password, str)
                and isinstance(other_sql_instance, bool)
                and isinstance(sql_instance_name, str)
                and isinstance(export_location, str)
                and isinstance(sql_user_name, str)
                and isinstance(sql_password, str)
                and isinstance(database, str)
                and isinstance(capture_ma, bool)
                and isinstance(capture_schedules, bool)
                and isinstance(capture_activity_control, bool)
                and isinstance(capture_opw, bool)
                and isinstance(capture_holidays, bool)
                and isinstance(auto_pick_cluster, bool)
                and isinstance(cs_name, str)
                and isinstance(client_ids, list)):
            raise SDKException('CommCellMigration', '101')

        if path_type.lower() == 'local':
            self._path_type = 0
        elif path_type.lower() == 'network':
            self._path_type = 1
        else:
            raise SDKException('CommCellMigration', '104')

        if other_sql_instance:
            if sql_instance_name == "" or sql_user_name == "" or sql_password == "":
                raise SDKException('CommCellMigration', '103')
            sql_password = b64encode(sql_password.encode()).decode()

        common_options = {
            "otherSqlInstance": other_sql_instance,
            "pathType": self._path_type,
            "dumpFolder": export_location,
            "splitCSDB": 1,
            "sqlLinkedServer": {
                "sqlServerName": sql_instance_name,
                "sqlUserAccount": {
                    "userName": sql_user_name,
                    "password": sql_password
                }
            }
        }

        if self._path_type == 1:
            if network_user_name == "" or network_user_password == "":
                raise SDKException('CommCellMigration', '103')
            network_user_password = b64encode(network_user_password.encode()).decode()
            common_options["userAccount"] = {
                "password": network_user_password,
                "userName": network_user_name
            }

        export_json = {
            "taskInfo": {
                "task": {
                    "taskType": 1,
                    "isEditing": False,
                    "initiatedFrom": 2,
                    "policyType": 0,
                    "taskFlags": {
                        "disabled": False
                    }
                },
                "appGroup": {
                },
                "subTasks": [
                    {
                        "subTask": {
                            "subTaskType": 1,
                            "operationType": 4029
                        },
                        "options": {
                            "adminOpts": {
                                "ccmOption": {
                                    "commonOptions": common_options,
                                    "captureOptions": {
                                        "captureMediaAgents": capture_ma,
                                        "lastHours": 60,
                                        "remoteDumpDir": "",
                                        "remoteCSName": "",
                                        "captureSchedules": capture_schedules,
                                        "captureActivityControl": capture_activity_control,
                                        "captureOperationWindow": capture_opw,
                                        "captureHolidays": capture_holidays,
                                        "pruneExportedDump": False,
                                        "autopickCluster": auto_pick_cluster,
                                        "copyDumpToRemoteCS": False,
                                        "useJobResultsDirForExport": False,
                                        "captureFromDB": {
                                            "csName": cs_name,
                                            "csDbName": database
                                        },
                                        "entities": [
                                        ],
                                        "timeRange": {
                                            "_type_": 54,
                                        }
                                    }
                                }
                            }
                        }
                    }
                ]
            }
        }

        if not other_sql_instance:
            del export_json['taskInfo']['subTasks'][0]['options']['adminOpts']['ccmOption'] \
                ['captureOptions']['captureFromDB']

        sub_dict = export_json['taskInfo']['subTasks'][0]['options']['adminOpts']['ccmOption'] \
            ['captureOptions']['entities']

        if other_entities:
            for entity in other_entities:
                if entity == "schedule_policies":
                    sub_dict.append({'commCellName': self._commcell_name, "_type_": 34})

                elif entity == "users_and_user_groups":
                    sub_dict.append({'commCellName': self._commcell_name, "_type_": 36})

                elif entity == "alerts":
                    sub_dict.append({'commCellName': self._commcell_name, "_type_": 42})

        if client_list:
            if other_sql_instance:
                if not sql_instance_name \
                        or not sql_user_name \
                        or not sql_password \
                        or not client_ids:
                    raise SDKException('CommCellMigration', '106')

                for index, client in enumerate(client_list):
                    temp_dic = {'clientName': client, "clientId": client_ids[index]}
                    sub_dict.append(temp_dic)

            else:
                exportable_clients = list(self._commcell_object.grc.get_clients_for_migration(
                    podcell_id=2, podcell_guid=self._commcell_object.commserv_guid
                ).values())
                for client in client_list:
                    if not self._commcell_object.clients.has_client(client):
                        raise SDKException(
                            'CommCellMigration', '107',
                            f'Client {client} not found'
                        )
                    agents = self._commcell_object.clients.get(client).agents.all_agents
                    if not agents:
                        raise SDKException(
                            'CommCellMigration', '107',
                            f'Client {client} does not have any agents'
                        )
                    temp_dic = {'clientName': client, 'commCellName': self._commcell_name}
                    sub_dict.append(temp_dic)

        flag, response = self._cvpysdk_object.make_request('POST',
                                                           self._services['RESTORE'],
                                                           export_json)

        if flag:
            if response.json() and 'jobIds' in response.json():
                return Job(self._commcell_object, response.json()['jobIds'][0])
            elif response.json() and 'errorCode' in response.json():
                raise SDKException('CommCellMigration', '102', 'CCM Export job failed with error code : ' +
                                   str(response.json()['errorCode']))
        else:
            response_string = self._update_response_(response.text)
            raise SDKException('Response', '101', response_string)

    def commcell_import(self, import_location, options_dictionary):
        """ Starts the Commcell Import job.

            Args:
                import_location     ( str )         --  Location to import the generated dumps.

                options_dictionary  ( dict )        --  Contains list of options used for CCMImport and default values.
                    {
                        "pathType": "Network",
                        "userName" : "username",
                        "password": "password",
                        "forceOverwrite": False,
                        "failIfEntityAlreadyExists": False,
                        "deleteEntitiesNotPresent": False,
                        "deleteEntitiesIfOnlyfromSource": False,
                        "forceOverwriteHolidays": False,
                        "mergeHolidays": True,
                        "forceOverwriteOperationWindow": False,
                        "mergeOperationWindow": False,
                        "forceOverwriteSchedule": False,
                        "mergeSchedules": True
                    }

            Returns:
                CCM Import Job instance             --  returns the CCM Import job instance.

            Raises:
                SDKException:
                    if type of the input is not valid.

                    if all the required inputs are not provided.

                    if invalid inputs are passed.
        """
        path_type = options_dictionary.get("pathType", "Local")
        network_user_name = options_dictionary.get("userName", "")
        network_user_password = options_dictionary.get("password", "")
        force_overwrite = options_dictionary.get('forceOverwrite', False)
        fail_if_entry_already_exists = options_dictionary.get('failIfEntityAlreadyExists', False)
        delete_entities_not_present = options_dictionary.get('deleteEntitiesNotPresent', False)
        delete_only_source = options_dictionary.get('deleteEntitiesIfOnlyfromSource', False)
        fo_holidays = options_dictionary.get("forceOverwriteHolidays", False)
        merge_holidays = options_dictionary.get("mergeHolidays", True)
        fo_operation_window = options_dictionary.get("forceOverwriteOperationWindow", False)
        merge_operation_window = options_dictionary.get("mergeOperationWindow", False)
        fo_schedules = options_dictionary.get("forceOverwriteSchedule", False)
        merge_schedules = options_dictionary.get("mergeSchedules", True)

        if not (isinstance(path_type, str) and isinstance(import_location, str)):
            raise SDKException('CommCellMigration', '101')

        common_options = {
            "bRoboJob": False,
            "databaseConfiguredRemote": False,
            "pathType": self._path_type,
            "dumpFolder": import_location,
            "splitCSDB": 0
        }

        if path_type.lower() == 'local':
            self._path_type = 0
        elif path_type.lower() == 'network':
            self._path_type = 1
            common_options["userAccount"] = {
                "password": network_user_password,
                "userName": network_user_name
            }
        else:
            raise SDKException('CommCellMigration', '104')

        if self._path_type == 1:
            if network_user_name == "" or network_user_password == "":
                raise SDKException('CommCellMigration', '103')

        import_json = {
            "taskInfo": {
                "associations": [
                    {
                        "type": 0,
                        "clientSidePackage": True,
                        "consumeLicense": True
                    }
                ],
                "task": {
                    "taskType": 1,
                    "initiatedFrom": 2,
                    "taskFlags": {
                        "disabled": False
                    }
                },
                "subTasks": [
                    {
                        "subTask": {
                            "subTaskType": 1,
                            "operationType": 4030
                        },
                        "options": {
                            "adminOpts": {
                                "ccmOption": {
                                    "mergeOptions": {
                                        "deleteEntitiesIfOnlyfromSource": False,
                                        "forceOverwriteHolidays": fo_holidays,
                                        "reuseTapes": False,
                                        "specifyStagingPath": False,
                                        "forceOverwriteOperationWindow": fo_operation_window,
                                        "fallbackSpareGroup": "",
                                        "mergeOperationWindow": merge_operation_window,
                                        "pruneImportedDump": False,
                                        "alwaysUseFallbackDataPath": True,
                                        "deleteEntitiesNotPresent": delete_entities_not_present,
                                        "deleteEntitiesIfOnlyfromSource": delete_only_source,
                                        "forceOverwrite": force_overwrite,
                                        "mergeHolidays": merge_holidays,
                                        "forceOverwriteSchedule": fo_schedules,
                                        "fallbackDrivePool": "",
                                        "mergeActivityControl": True,
                                        "fallbackMediaAgent": "",
                                        "mergeSchedules": merge_schedules,
                                        "failIfEntityAlreadyExists": fail_if_entry_already_exists,
                                        "fallbackLibrary": "",
                                        "skipConflictMedia": False,
                                        "stagingPath": ""
                                    },
                                    "commonOptions": common_options
                                }
                            }
                        }
                    }
                ]
            }
        }
        flag, response = self._cvpysdk_object.make_request('POST',
                                                           self._services['RESTORE'],
                                                           import_json)

        if flag:
            if response.json() and 'jobIds' in response.json():
                return Job(self._commcell_object, response.json()['jobIds'][0])
            elif response.json() and 'errorCode' in response.json():
                raise SDKException('CommCellMigration', '102', 'CCM Import job failed with error code : ' +
                                   str(response.json()['errorCode']))
        else:
            response_string = self._update_response_(response.text)
            raise SDKException('Response', '101', response_string)

    def tape_import(self, library_id, medias_id, drive_pool_id):

        """ performs the tape import import operation for the specified tape.

            Args:
                library_id      (int)       --      tape library id.

                medias_id        (list)       --      tape id.

                drive_pool_id   (int)       --      drive pool id

            Returns:
                Tape import job instance
        """

        tape_import_json = {
            "taskInfo": {
                "associations": [
                ], "task": {
                    "ownerId": 1, "taskType": 1, "ownerName": "admin", "sequenceNumber": 0, "initiatedFrom": 1,
                    "policyType": 0, "taskId": 0, "taskFlags": {
                        "disabled": False
                    }
                }, "subTasks": [
                    {
                        "subTask": {
                            "subTaskType": 1, "operationType": 4017
                        },
                        "options": {
                            "adminOpts": {
                                "contentIndexingOption": {
                                    "subClientBasedAnalytics": False
                                }, "libraryOption": {
                                    "operation": 15, "media": [
                                    ], "library": {
                                        "libraryName": "", "_type_": 9, "libraryId": library_id
                                    }, "catalogMedia": {
                                        "fileMarkerToStart": 0, "fileMarkerToEnd": 0, "reCatalog": True,
                                        "maxNumOfDrives": 1,
                                        "spareGroupId": 0,
                                        "merge": True,
                                        "subTaskType": 2,
                                        "drivePoolEntity": {
                                            "_type_": 47, "drivePoolId": drive_pool_id
                                        }
                                    }, "mediaAgent": {
                                        "mediaAgentId": 2, "_type_": 11
                                    }
                                }
                            }, "restoreOptions": {
                                "virtualServerRstOption": {
                                    "isBlockLevelReplication": False
                                }, "commonOptions": {
                                    "syncRestore": False
                                }
                            }
                        }
                    }
                ]
            }
        }

        sub_dict = tape_import_json["taskInfo"]["subTasks"][0]["options"]["adminOpts"]["libraryOption"]["media"]

        for media in medias_id:
            temp_dict = {"_type_": 46, "mediaId": int(media), "mediaName": ""}
            sub_dict.append(temp_dict)

        flag, response = self._cvpysdk_object.make_request('POST',
                                                           self._services['RESTORE'],
                                                           tape_import_json)

        if flag:
            if response.json() and 'jobIds' in response.json():
                return Job(self._commcell_object, response.json()['jobIds'][0])
            elif response.json() and 'errorCode' in response.json():
                raise SDKException('CommCellMigration', '102', 'Tape Import job failed with error code : ' +
                                   str(response.json()['errorCode']))
        else:
            response_string = self._update_response_(response.text)
            raise SDKException('Response', '101', response_string)

class GlobalRepositoryCell:
    """Class for representing the GRC feature from commcell"""

    def __init__(self, commcell_object):
        """
        Initializes the object of GlobalRepositoryCell class

        Args:
            commcell_object (Commcell)  -   Commcell class instance

        Returns:
            grc (GlobalRepositoryCell) - instance of the GlobalRepositoryCell class
        """
        self._commcell_object = commcell_object
        self._cvpysdk_object = self._commcell_object._cvpysdk_object
        self._services = self._commcell_object._services
        self._commcell_name = self._commcell_object.commserv_name

    def _get_task_details(self, task_id):
        """
        Util for getting XML of GRC schedule task (required for generating more XMLs)

        Args:
            task_id     (int)   -   id of grc schedule's task

        Returns:
            task_xml    (str)   -   xml form string with grc schedule details
            Example:
                <TMMsg_GetTaskDetailResp>
	                <taskInfo>
		                <task taskId="" taskName="" > ... </task>
                        <appGroup/>
                        <subTasks>
                            <subTask subTaskId="" subTaskType="" ...>
                            <options>
                                <backupOpts backupLevel="">
                                    <dataOpt autoCopy=""/>
                                </backupOpts>
                                <adminOpts>
                                    <ccmOption>
                                        <mergeOptions ...>
                                        <captureOptions ...>
                                            ...
                                        </captureOptions>
                                    </ccmOption>
                                </adminOpts>
                            </options>
                            <pattern ...>...</pattern>
                        </subTasks>
                    </taskInfo>
                </TMMsg_GetTaskDetailResp>
        """
        get_task_xml = f'<TMMsg_GetTaskDetailReq taskId="{task_id}"/>'
        return self._commcell_object.qoperation_execute(get_task_xml, return_xml=True)

    def _get_commcell_from_id(self, commcell_id):
        """
        Util to get registered commcell name from given commcell id

        Args:
            commcell_id (int)   -   id of commcell

        Returns:
            commcell_name   (str)   -   name of commcell
        """
        for commcell_name, commcell_data in self._commcell_object.registered_commcells.items():
            if commcell_data.get('commCell', {}).get('commCellId') == commcell_id:
                return commcell_name

    def _modify_task_props(self, podcell_properties, task_xml):
        """
        Util for modifying task properties, after grc properties are updated

        Args:
            podcell_properties  (dict)  -   the dict returned by get_podcell_properties
            task_xml    (str)           -   the xml returned for grc schedule's task info

        Returns:
            response    (dict)   -   the response from execute qoperation
        """
        grc_schedule_xml = ET.fromstring(podcell_properties['schedule_xml'])
        task_info_xml = ET.fromstring(task_xml)
        modify_task_xml = """
        <TMMsg_ModifyTaskReq>
            <taskInfo>
                <task initiatedFrom="1" ownerId="{0}" ownerName="{1}" policyType="0" sequenceNumber="0" taskId="{2}" taskType="2">
                    <taskFlags disabled="0" isEZOperation="0" isEdgeDrive="0"/>
                </task>
                <appGroup/>
                {3}
            </taskInfo>
        </TMMsg_ModifyTaskReq>
        """
        modify_task_xml = modify_task_xml.format(
            grc_schedule_xml.find('taskInfo/task').get('ownerId'),
            grc_schedule_xml.find('taskInfo/task').get('ownerName'),
            podcell_properties['task_id'],
            ET.tostring(task_info_xml.find('taskInfo/subTasks'), encoding='unicode')
        )
        return self._commcell_object.qoperation_execute(modify_task_xml)

    def _get_podcell_entities(self, podcell_name: str = None, podcell_id: int = None, podcell_guid: str = None):
        """
        Gets the entities in podcell available for monitoring via GRC

        Args:
            podcell_name    (str)   -   name of pod cell
            podcell_id      (int)   -   id of podcell
            podcell_guid    (str)   -   guid of podcell (Optional)

        Returns:
            monitor_entities    (str)   -   all entities of pod cell in XML format
            Example:
                <EVGui_CCMCommCellInfo commcellName="" commcellNumber="" commcellId="">
                    <clientEntityLst clientId="" clientName="">
                        ...
                    </clientEntityLst>
                    <clientEntityLst clientId="" clientName="">
                        ...
                    </clientEntityLst>
                    <clientEntityLst clientId="" ...>
                        <appTypeEntityList ... appTypeId="">
                            <instanceList ... instanceId="">
                                <backupSetList ... backupsetId="">
                                    <subclientList ... subclientId=""/>
                                </backupSetList>
                            </instanceList>
                        </appTypeEntityList>
                    </clientEntityLst>
                    <clientComputerGrp clientGroupId="" clientGroupName=""/>
                    ...
                    <clientComputerGrp clientGroupId="" clientGroupName=""/>
                </EVGui_CCMCommCellInfo>
        """
        if podcell_id is None:
            if podcell_name is None:
                raise SDKException('GlobalRepositoryCell', '103')
            podcell_id = self._commcell_object.registered_commcells.get(podcell_name, {}) \
                .get('commCell', {}).get('commCellId')
            if podcell_id is None:
                raise SDKException('GlobalRepositoryCell', '104', f'for podcell: {podcell_name}')
        if podcell_name is None:
            podcell_name = self._get_commcell_from_id(podcell_id)
        if podcell_guid is None:
            podcell_guid = self._commcell_object.registered_commcells[podcell_name].get('commCell', {}).get('csGUID')

        entities_xml = """
        <EVGui_GetCCMExportInfo exportMsgType="3" strCSName="{0}*{0}*8400">
            <mediaAgent _type_="3"/>
            <userInfo/>
            <commCell _type_="1" commCellId="{1}" commCellName="{0}" csGUID="{2}"/>
        </EVGui_GetCCMExportInfo>
        """
        exec_xml = entities_xml.format(podcell_name, podcell_id, podcell_guid)
        resp = self._commcell_object.qoperation_execute(exec_xml)
        return resp.get('strXmlInfo')

    def get_clients_for_migration(self, podcell_name: str = None, podcell_id: int = None, podcell_guid: str = None):
        """
        Gets the podcell clients that can be migrated
        
        Args:
            podcell_name    (str)   -   name of pod cell
            podcell_id      (int)   -   id of podcell
            podcell_guid    (str)   -   guid of podcell (Optional)
        
        Returns:
            clients_dict    (dict)  -   dict with client ID as key and client name value
            Example:
                {
                    X: "clientA",
                    Y: "clientB",
                    Z: "clienta"
                }
        """
        clients_dict = {}
        entities_xml = self._get_podcell_entities(
            podcell_name=podcell_name,
            podcell_id=podcell_id,
            podcell_guid=podcell_guid
        )
        entities_xml = ET.fromstring(entities_xml)
        for client_node in entities_xml.findall('clientEntityLst'):
            cl_id = client_node.get('clientId')
            cl_name = client_node.get('clientName')
            clients_dict[cl_id] = cl_name
        return clients_dict

    def _get_podcell_properties(self, podcell_name: str = None, podcell_id: int = None):
        """
        Gets the GRC properties of given pod cell

        Args:
            podcell_name    (str)   -   name of pod cell
            podcell_id      (int)   -   id of podcell

        Returns:
            podcell_properties  (dict)  -   different properties of pod cell in dict format with xml values
        """
        # TODO: Update grc properties map
        grc_prop_map = {
            2: 'podcell_name',
            4: 'schedule_xml',
            15: 'entities_xml',
            16: 'libraries_xml',
            19: 'task_id'
        }
        if podcell_id is None:
            if podcell_name is None:
                raise SDKException('GlobalRepositoryCell', '103')
            podcell_id = self._commcell_object.registered_commcells.get(podcell_name, {}) \
                .get('commCell', {}).get('commCellId')
            if podcell_id is None:
                raise SDKException('GlobalRepositoryCell', '104', f'for podcell: {podcell_name}')
        grc_props_xml = f'<App_GetGRCCommCellPropsReq commcellId="{podcell_id}"/>'
        grc_props_response = self._commcell_object.qoperation_execute(grc_props_xml)
        podcell_properties = {
            grc_prop_map.get(prop.get('propId'), prop.get('propId')): prop.get('stringVal') or prop.get('numVal')
            for prop in grc_props_response['grcCommcellPropList']
        }
        return podcell_properties

    def modify_monitored_clients(self, podcell_name: str = None, podcell_id: int = None, clients: list = None):
        """
        Modifies (overwrites) the monitored clients in grc properties for given podcell

        Args:
            podcell_name    (str)   -   name of pod cell
            podcell_id      (int)   -   id of podcell
            client_ids      (list)  -   list of client ids, names
                                        or Client objects (of pod cell)

        Returns:
            None
        """
        if podcell_id is None:
            if podcell_name is None:
                raise SDKException('GlobalRepositoryCell', '103')
            podcell_id = self._commcell_object.registered_commcells.get(podcell_name, {}) \
                .get('commCell', {}).get('commCellId')
            if podcell_id is None:
                raise SDKException('GlobalRepositoryCell', '104', f'for podcell: {podcell_name}')

        set_grc_xml = """
            <App_SetGRCCommCellPropsReq commcellId="{0}">
                <grcCommcellProp numVal="0" propId="4" stringVal="{1}"/>
                <grcCommcellProp numVal="1" propId="1" stringVal=""/>
                <grcCommcellProp numVal="0" propId="2" stringVal="{2}"/>
                <grcCommcellProp numVal="0" propId="15" stringVal="{3}"/>
                <grcCommcellProp numVal="1" propId="8" stringVal=""/>
                <grcCommcellProp numVal="0" propId="14" stringVal=""/>
            </App_SetGRCCommCellPropsReq>
        """
        xml_header = '<?xml version=\'1.0\' encoding=\'UTF-8\'?>'
        cc_props = self._get_podcell_properties(podcell_id=podcell_id)
        podcell_name = cc_props['podcell_name']
        task_xml = self._get_task_details(task_id=cc_props['task_id'])
        podcell_entities = self._get_podcell_entities(podcell_id=podcell_id)
        entities_xml = ET.fromstring(podcell_entities)
        client_ids = []
        if isinstance(clients[0], str):
            for client_node in entities_xml.findall('clientEntityLst'):
                if client_node.get('clientName') in clients:
                    client_ids.append(client_node.get('clientId'))
        elif isinstance(clients[0], int):
            client_ids = clients
        elif isinstance(clients[0], Client):
            client_ids = [int(cl.client_id) for cl in clients]

        # Generate nested XML 1 (selected clients)
        current_schedule = ET.fromstring(cc_props['schedule_xml'])
        capture_options = current_schedule.find('taskInfo/subTasks/options/adminOpts/ccmOption/captureOptions')
        # remove all <entities ...> tags
        for entity_node in capture_options.findall('entities'):
            capture_options.remove(entity_node)
        # insert <entities ...> tags for selected client_ids
        for clid in client_ids:
            capture_options.insert(0, ET.Element('entities', {'clientId': str(clid), '_type_': '3'}))
        nested_xml1 = ET.tostring(current_schedule, encoding='unicode')
        nested_xml1 = html.escape(f'{xml_header}{nested_xml1}')

        # Generate nested XML 2 (all clients in podcell)
        entities_xml = ET.fromstring(podcell_entities)
        nested_xml2 = ET.tostring(entities_xml, encoding='unicode')
        nested_xml2 = html.escape(nested_xml2)

        # Combine nested XMLs into parent XML
        final_xml = set_grc_xml.format(podcell_id, nested_xml1, podcell_name, nested_xml2)
        self._commcell_object.qoperation_execute(final_xml)
        self._modify_task_props(cc_props, task_xml)



# -*- coding: utf-8 -*-

# --------------------------------------------------------------------------
# Copyright Commvault Systems, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# --------------------------------------------------------------------------

""" File for operating on a cloud database instance.

AzureCosmosDBInstance:   Derived class from CloudAppsInstance Base class, representing a
                        Azure Cosmos DB instance( Azure CosmosDB Cassandra API, ..), and to
                        perform operations on that instance

AzureCosmosDBInstance:

    __init__()                      --  Initializes Azure Cosmos DB instance object with associated
    agent_object, instance name and instance id

    _get_instance_properties()      --  Retrieves cloud database related instance properties

    restore()                       -- Submits a restore request based on restore options

"""

from __future__ import unicode_literals
import time
from ..cainstance import CloudAppsInstance
from ...exception import SDKException


class AzureCosmosDBInstance(CloudAppsInstance):
    """
    Class for representing an Instance of the Azure Cosmos DB instance such as
    Azure CosmosDB Cassandra API

    """

    def __init__(self, agent_object, instance_name, instance_id=None):
        """Initializes the object of the AzureCosmosDBInstance class

            Args:
                agent_object    (object)  --  instance of the Agent class

                instance_name   (str)     --  name of the instance

                instance_id     (str)     --  id of the instance
                    default: None

            Returns:
                object - instance of the Instance class

        """
        self._agent_object = agent_object
        self._ca_instance_type = None
        self._browse_request = {}
        self._browse_url = None

        super(
            AzureCosmosDBInstance,
            self).__init__(
                agent_object,
                instance_name,
                instance_id)

    def restore(self, restore_options):
        """
            Restores the content of this instance content

            Args:
                restore_options  : dict of keyword arguments as follows:
                    Example:
                       restore_options = {
                            'no_of_streams': no_of_stream,
                            'destination_instance': 'destination_instance_name',
                            'destination_instance_id': destination_instance_id,
                            'paths': ['path_to_restored'],
                            'cloudinstancetype': 'cloudinstancetype,
                            'backupsetname': 'backupsetname',
                            'unconditional_overwrite': True,
                            'in_place': True,
                            'sourcedatabase': 'sourcekeyspacename',
                            'destinationdatabase': 'destinationkeyspacename',
                            'srcstorageaccount': 'sourcestorageaccount',
                            'deststorageaccount': 'destinationstorageaccount'
                        }

            Returns:

                object - instance of the Job class for this restore job
        """
        if not (isinstance(restore_options, dict)):
            raise SDKException('Instance', '101')
        request_json = self._restore_json(restore_option=restore_options)

        request_json["taskInfo"]["associations"][0]["_type_"] = "INSTANCE_ENTITY"
        request_json["taskInfo"]["associations"][0]["cloudInstanceType"] = restore_options["cloudinstancetype"]
        request_json["taskInfo"]["associations"][0]["backupsetName"] = restore_options["backupsetname"]
        request_json["taskInfo"]["subTasks"][0]["options"]["restoreOptions"]

        cloud_app_restore_options = {
            "azureDbRestoreOptions": {
                "overwrite": restore_options["unconditional_overwrite"],
                "restoreEntity": [
                ]
            },
            "instanceType": restore_options["cloudinstancetype"]
        }

        if "CASSANDRA" in restore_options["cloudinstancetype"]:
            if restore_options.get("tempWriteThroughput", 0):
                cloud_app_restore_options["azureDbRestoreOptions"]["tempWriteThroughput"] = restore_options["tempWriteThroughput"]
            cloud_app_restore_options["azureDbRestoreOptions"]["restoreEntity"] = [
                {
                    "srcEntity": {
                        "databaseName": restore_options["sourcedatabase"],
                        "storageAccountName": restore_options["srcstorageaccount"]},
                    "destEntity": {
                        "databaseName": restore_options["destinatinodatabase"],
                        "storageAccountName": restore_options["deststorageaccount"]}}]

        request_json["taskInfo"]["subTasks"][0]["options"]["restoreOptions"]["cloudAppsRestoreOptions"] = cloud_app_restore_options

        return self._process_restore_response(request_json)



# -*- coding: utf-8 -*-
# --------------------------------------------------------------------------
# Copyright Commvault Systems, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# --------------------------------------------------------------------------
""" File for operating on a yugabyte instance.
YugabyteInstance:   Derived class from BigDataAppsInstance Base class, representing a
                        YugabyteDB instance, and to perform operations on that instance
YugabyteInstance:
    __init__()                      --  Initializes YugabyteDB instance object with associated
    agent_object, instance name and instance id
    restore()                       -- Submits a restore request based on restore options
"""
from __future__ import unicode_literals
import time
from ..bigdataappsinstance import BigDataAppsInstance
from ...exception import SDKException


class YugabyteInstance(BigDataAppsInstance):
    """
    Class for representing an Instance of the yugabyte instance
    """
    def __init__(self, agent_object, instance_name, instance_id=None):
        """Initializes the object of the YugabyteInstance class
            Args:
                agent_object    (object)  --  instance of the Agent class
                instance_name   (str)     --  name of the instance
                instance_id     (str)     --  id of the instance
                    default: None
            Returns:
                object - instance of the Instance class
        """
        self._agent_object = agent_object
        self._browse_request = {}
        self._browse_url = None
        super(
            YugabyteInstance,
            self).__init__(
                agent_object,
                instance_name,
                instance_id)

    def restore(self, restore_options):
        """
            Restores the content of this instance content
            Args:
                restore_options : dict of keyword arguments needed to submit a yugabytedb restore:

                    Example:
                       restore_options = {
                            'no_of_streams': no_of_stream,
                            'multinode_restore': multinode_restore,
                            'destination_instance': 'destination_instance_name',
                            'destination_instance_id': destination_instance_id,
                            'paths': ['paths_to_be_restored'],
                            'destination_client_id': 'destination_client_id,
                            'destination_client_name': 'destination_client_name',
                            'client_type': client_type,
                            'destination_appTypeId': destination_appTypeId,
                            'backupset_name': 'backupset_name',
                            'sql_fromtable': 'sql_fromtable',
                            'cql_fromtable': 'cql_fromtable',
                            'sql_totable': 'sql_totable',
                            'cql_totable': 'cql_totable',
                            'accessnodes': ['accessnodes'],
                            'kms_config': 'kms_config',
                            'kmsconfigUUID': 'kmsconfigUUID
                        }
            Returns:
                object - instance of the Job class for this restore job
        """
        if not (isinstance(restore_options, dict)):
            raise SDKException('Instance', '101')
        request_json = self._restore_json(restore_option=restore_options)
        request_json["taskInfo"]["associations"][0]["subclientId"] = restore_options.get(
            "subclient_id", -1)
        request_json["taskInfo"]["associations"][0]["backupsetName"] = restore_options.get(
            "backupset_name")

        distributed_restore_json = {
            "distributedRestore": True,
        }

        request_json["taskInfo"]["subTasks"][0]["options"]["restoreOptions"]["cloudAppsRestoreOptions"] = {
            "instanceType": "AMAZON_S3"
        }
        distributed_restore_json["noSQLGenericRestoreOptions"] = {
            "tableMap": [
                {
                    "fromTable": restore_options.get("sql_fromtable"),
                    "toTable": restore_options.get("sql_totable")
                },
                {
                    "fromTable": restore_options.get("cql_fromtable"),
                    "toTable": restore_options.get("cql_totable")
                }
            ]
        }
        distributed_restore_json["yugabyteDBRestoreOptions"] = {
            "kmsConfigName": restore_options.get("kms_config"),
            "kmsConfigUUID": restore_options.get("kmsconfigUUID")
        }
        distributed_restore_json["clientType"] = restore_options.get("client_type")
        distributed_restore_json["isMultiNodeRestore"] = True
        access_nodes = []
        for node in restore_options.get("accessnodes"):
            client_object = self._commcell_object.clients.get(node)
            client_id = int(client_object.client_id)
            access_node = {
                "clientId": client_id,
                "clientName": client_object.client_name
            }
            access_nodes.append(access_node)
        distributed_restore_json["dataAccessNodes"] = {
            "dataAccessNodes": access_nodes
        }

        request_json["taskInfo"]["subTasks"][0]["options"]["restoreOptions"][
            "distributedAppsRestoreOptions"] = distributed_restore_json
        return self._process_restore_response(request_json)



# -*- coding: utf-8 -*-

# --------------------------------------------------------------------------
# Copyright Commvault Systems, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# --------------------------------------------------------------------------

""" File for associating Workload and Backup destination regions for various entites
class: Regions. Region

Regions:
========
    _get_regions()                  --  Gets all the regions created in commcell
    refresh()                       --  Refresh the list of Regions associated to commcell
    has_region()                    --  Checks if region with given name exists
    get()                           --  returns Region class object for the specified input name
    set_region()                    --  Associate a region to an entity
    get_region()                    --  Gets the Region associated to an Entity
    calculate_region()              --  Calculates the Region to be associated to an Entity
    add()                           --  Method to add a region
    delete()                        --  Method to delete regions

Attributes:

    ***all_regions***               --  returns dict of details about region such as id

Region:
=======
    _get_region_id()                --  Returns the region id
    _get_region_properties()        --  Gets the properties of this region
    refresh()                       --  Refresh the properties of the regions

Attributes:
    ***region_id***                 --  Id of the given Region
    ***region_name***               --  Name of the given region
    ***region_type***               --  type of the given region
    ***locations***                 --  locations of the given region
    ***associated_servers_count***  --  associated servers count of the given region
    ***associated_servers***        --  associated servers of the given region
    ***associated_plans_count***    --  associated plans count of the given region
    ***associated_plans***          --  associated plan of the given region

"""
from .exception import SDKException


class Regions:
    """
    Class for representing all the Regions created in the commcell
    """
    def __init__(self, commcell_object):
        """Initialises the object of Regions class"""
        self._commcell_object = commcell_object
        self._cvpysdk_object = commcell_object._cvpysdk_object
        self._update_response_ = commcell_object._update_response_
        self._regions_api = self._commcell_object._services['REGIONS']
        self._EDIT_REGION = self._commcell_object._services['EDIT_REGION']
        self._GET_REGION = self._commcell_object._services['GET_REGION']
        self._CALCULATE_REGION = self._commcell_object._services['CALCULATE_REGION']
        self._regions = None
        self.refresh()

    def _get_regions(self):
        """Gets all the regions created in commcell"""
        flag, response = self._cvpysdk_object.make_request('GET', self._regions_api)
        if flag:
            regions = {}
            if response.json() and 'regions' in response.json():

                name_count = {}

                for region in response.json()['regions']:
                    temp_name = region['name'].lower()
                    temp_company = region.get('company', {}).get('name', '').lower()

                    if temp_name in name_count:
                        name_count[temp_name].add(temp_company)
                    else:
                        name_count[temp_name] = {temp_company}

                for region in response.json()['regions']:
                    temp_name = region['name'].lower()
                    temp_id = region['id']
                    temp_company = region.get('company', {}).get('name', '').lower()

                    if len(name_count[temp_name]) > 1:
                        unique_key = f"{temp_name}_({temp_company})"
                    else:
                        unique_key = temp_name
                    regions[unique_key] = temp_id

                return regions

            raise SDKException('Response', '102')

        response_string = self._update_response_(response.text)
        raise SDKException('Response', '101', response_string)

    def refresh(self):
        """Refresh the list of Regions associated to commcell"""
        self._regions = self._get_regions()

    def has_region(self, name):
        """Checks if the given Region exists in the Commcell.

            Args:
                name    (str)   --  name of the Region

            Returns:
                bool    -   boolean output whether the Region exists in the commcell or not

            Raises:
                SDKException:
                    if type of the Region name argument is not string

        """
        if not isinstance(name, str):
            raise SDKException('Region', '103')

        return self._regions and (name.lower() in self._regions)

    def get(self, name):
        """
        Returns the instance of Region class for the given Region name
        Args:
            name    (str)   --  name of the Region

        Returns:
            object  -- Instance of Region class for the given Region name

        Raises:
            SDKException:
                - If the Region name argument is not a string

                - If No Region found in commcell with the given region name
        """
        if not isinstance(name, str):
            raise SDKException('Region', '102', "Invalid input received")

        if not self.has_region(name):
            raise SDKException('Region', '102', f"Region {name.lower()} not present in commcell")

        return Region(self._commcell_object, name, self._regions[name.lower()])

    def set_region(self, entity_type, entity_id, entity_region_type, region_id):
        """
        Associate a region to an entity
        Args:
            entity_type         (str)   :   Type of the entity
                                            (eg:    COMMCELL,
                                                    COMPANY,
                                                    CLIENT,
                                                    CLIENT_GROUP,
                                                    MEDIAAGENT,
                                                    STORAGE_POOL, etc
                                            )
            entity_id           (int/str):   unique id of the entity

            entity_region_type  (str)   :   Type of the region
                                            (WORKLOAD or BACKUP)

            region_id           (int)   :   ID of the region from app_regions
        """
        if isinstance(region_id,str):
            region_id = int(region_id)
        request = {
            "entityRegionType": entity_region_type,
            "region":
                {
                    "id": region_id
                }
        }

        flag, response = self._commcell_object._cvpysdk_object.make_request(
            'PUT', self._EDIT_REGION % (entity_type, entity_id), request
        )

        if flag:
            if response.json():
                error_code = response.json()['errorCode']

                if error_code != 0:
                    if error_code == 50000:
                        raise SDKException('Regions', '101')
                    elif error_code == 547:
                        raise SDKException('Regions', '102', 'Invalid regionID provided in request')
                    else:
                        error_string = response.json()['errorMessage']
                        raise SDKException('Regions', '102', '{0}'.format(error_string))

            else:
                raise SDKException('Response', '102')
        else:
            response_string = self._commcell_object._update_response_(response.text)
            raise SDKException('Response', '101', response_string)

    def get_region(self, entity_type, entity_id, entity_region_type):
        """
        Gets the Region associated to an Entity
        Args:
            entity_type         (str)   :   Type of the entity
                                            (eg:    COMMCELL,
                                                    COMPANY,


                                                    CLIENT,
                                                    CLIENT_GROUP,
                                                    MEDIAAGENT,
                                                    STORAGE_POOL, etc
                                            )
            entity_id           (int)   :   unique id of the entity

            entity_region_type  (str)   :   Type of the region
                                            (WORKLOAD or BACKUP)
        """
        flag, response = self._commcell_object._cvpysdk_object.make_request(
            'GET', self._GET_REGION % (entity_type, entity_id, entity_region_type)
        )

        if flag:
            if response.json():
                try:
                    if response.json()['errorCode']:
                        error_string = response.json()['errorMessage']
                        raise SDKException('Regions', '102', '{0}'.format(error_string))

                except:
                    if response.json().get('regionId'):
                        return response.json().get('regionId')
                    return 0

            else:
                return None
        else:
            response_string = self._commcell_object._update_response_(response.text)
            raise SDKException('Response', '101', response_string)

    def calculate_region(self, entity_type, entity_id, entity_region_type):
        """
                Calculates the Region to be associated to an Entity
                Args:
                    entity_type         (str)   :   Type of the entity
                                                    (eg:    COMMCELL,
                                                            COMPANY,
                                                            CLIENT,
                                                            CLIENT_GROUP,
                                                            MEDIAAGENT,
                                                            STORAGE_POOL, etc
                                                    )
                    entity_id           (int)   :   unique id of the entity

                    entity_region_type  (str)   :   Type of the region
                                                    (WORKLOAD or BACKUP)
                """
        flag, response = self._commcell_object._cvpysdk_object.make_request(
            'GET', self._CALCULATE_REGION % (entity_type, entity_id, entity_region_type)
        )

        if flag:
            if response.json():
                try:
                    if response.json()['errorCode']:
                        error_string = response.json()['errorMessage']
                        raise SDKException('Regions', '102', '{0}'.format(error_string))

                except:
                    return response.json()['regionId']

            else:
                raise SDKException('Response', '102')
        else:
            response_string = self._commcell_object._update_response_(response.text)
            raise SDKException('Response', '101', response_string)

    def add(self, region_name: str, region_type: str, locations: list) -> object:
        """
        Method to add a region

            Args:
                region_name    (str)      --      name of the region
                region_type    (str)      --      type of region
                locations      (list)     --      list of dictionaries containing details of a location
                    e.g. locations = [{"city":"city1",
                                        "state":"stateOfCity1",
                                        "country":"countryOfCity1",
                                        "latitude":"latitudeOfCity1",
                                        "longitude":"longitudeOfCity1"},
                                        {"city":"city2",
                                        "state":"stateOfCity2",
                                        "country":"countryOfCity2",
                                        "latitude":"latitudeOfCity2",
                                        "longitude":"longitudeOfCity2"}]]

            Returns:
                object  --   instance of the region class created by this method

            Raises:
                SDKException:
                    if input parameters are incorrect

                    if Plan already exists

                    if invalid region type is passed
        """
        valid_region_types = ['USER_CREATED', 'OCI', 'AWS', 'AZURE', 'GCP']

        if not (isinstance(region_name, str) and isinstance(region_type, str) and isinstance(locations, list)):
            raise SDKException('Region', '101')

        elif self.has_region(region_name):
            raise SDKException('Region', '102', 'Region "{0}" already exists'.format(region_name))

        elif region_type.upper() not in valid_region_types:
            raise SDKException('Region', '102', 'Invalid region type')

        request_json = {
            "name": region_name,
            "type": region_type,
            "locations": locations
        }
        flag, response = self._cvpysdk_object.make_request('POST', self._regions_api, request_json)

        if flag:
            if response.json():
                response_value = response.json()
                error_message = response_value.get('errorMessage')
                error_code = response_value.get('errorCode', 0)

                if error_code != 0:
                    raise SDKException('Region', '102', f'Failed to create new Region\nError: "{error_message}"')

                region_name = response_value['name']

                self.refresh()

                return self.get(region_name)
            else:
                raise SDKException('Response', '102')
        else:
            response_string = self._update_response_(response.text)
            raise SDKException('Response', '101', response_string)

    def delete(self, region_name: str) -> None:
        """
        Method to delete regions

            Args:
                region_name    (str)   --  name of the region

            Returns:
                None    --  if the region is removed successfully

            Raises:
                SDKException:
                    if type of the Region name argument is not string

                    if no region exists with the given name

                    if response is empty

                    if failed to delete the region

        """
        if not isinstance(region_name, str):
            raise SDKException('Region', '101')

        else:
            region_name = region_name.lower()

            if not self.has_region(region_name):
                raise SDKException('Region', '102', 'No Region exists with name: "{0}"'.format(region_name))

            region_id = self._regions[region_name]
            delete_region_service = self._commcell_object._services['REGION']%(region_id)
            flag, response = self._cvpysdk_object.make_request('DELETE', delete_region_service)

            if not flag:
                response_string = self._update_response_(response.text)
                raise SDKException('Response', '101', response_string)
            if not response.json():
                raise SDKException('Response', '102' 'Empty response received from server.')
            response_json = response.json()
            error_code = str(response_json.get('errorCode'))
            error_message = response_json.get('errorMessage')

            if error_code == '0':
                self.refresh()
            else:
                raise SDKException('Region', '102', f'Failed to delete region. Error: {error_message}')

    @property
    def all_regions(self):
        """Returns dict consisting of all regions details such as id"""
        return self._regions


class Region:
    """ Class for performing operations on a given Region """
    def __init__(self, commcell_object, region_name, region_id=None):
        """ Initialise the Region class instance.
            Args:
                commcell_object     (object)    --  instance of the Commcell class

                region_name         (str)       --  name of the region

                region_id           (int)       --  ID of the REgion
                                                    Default : None

            Returns:
                object  -   instance of the Region class
        """
        self._commcell_object = commcell_object
        self._region_name = region_name
        self._cvpysdk_object = commcell_object._cvpysdk_object
        self._update_response_ = commcell_object._update_response_

        self._region_name = region_name

        if region_id:
            self._region_id = str(region_id)
        else:
            self._region_id = self._get_region_id()

        self._region_api = self._commcell_object._services['REGION']%(self._region_id)
        self._region_type = None
        self._locations = []
        self._associated_servers_count = None
        self._associated_servers = []
        self._associated_plans_count = None
        self._associated_plans = []
        self.refresh()
        self._region_properties = None

    def _get_region_id(self):
        """ Returns the ID of the Region """
        regions = Regions(self._commcell_object)
        return regions.get(self._region_name).region_id

    def _get_region_properties(self):
        """Gets the properties of this region"""
        flag, response = self._commcell_object._cvpysdk_object.make_request('GET', self._region_api)

        if flag:
            if response.json():
                properties = response.json()
                self._region_id = properties.get('id')
                self._region_name = properties.get('name')
                self._region_type = properties.get('regionType')
                if 'locations' in properties:
                    self._locations = properties.get('locations')
                if 'associatedServers' in properties:
                    self._associated_servers_count = properties.get('associatedServers', {}).get('serversCount')
                    if 'servers' in properties.get('associatedServers'):
                        self._associated_servers = properties.get('associatedServers', {}).get('servers')
                if 'associatedRegionBasedPlans' in properties:
                    self._associated_plans_count = properties.get('associatedRegionBasedPlans', {}).get('plansCount')
                    if 'plans' in properties.get('associatedRegionBasedPlans'):
                        for plan in properties.get('associatedRegionBasedPlans', {}).get('plans'):
                            self._associated_plans.append(plan.get('plan'))
            else:
                raise SDKException('Response', '102')

        else:
            response_string = self._commcell_object._update_response_(response.text)
            raise SDKException('Response', '101', response_string)

    @property
    def region_id(self) -> int:
        """
        Get Region ID

        Returns:
            int -- region ID
        """
        return self._region_id

    @property
    def region_name(self) -> str:
        """
        Get Region name

        Returns:
            str -- name of the region
        """
        return self._region_name

    @property
    def region_type(self) -> str:
        """
        Get Region type

        Returns:
            str -- type of the region
        """
        return self._region_type

    @property
    def locations(self) -> list:
        """
        Get locations

        Returns:
            list -- locations added in the region
        """
        return self._locations

    @property
    def associated_servers_count(self) -> int:
        """
         Get associated servers count

        Returns:
            int -- count of servers associated with the region
         """
        return self._associated_servers_count

    @property
    def associated_servers(self) -> list:
        """
         Get associated servers

        Returns:
            list -- list of servers associated with the region
         """
        return self._associated_servers

    @property
    def associated_plans_count(self) -> int:
        """
        Get associated plans count

        Returns:
            int -- count of plans associated with the region
        """
        return self._associated_plans_count

    @property
    def associated_plans(self) -> list:
        """
        Get associated plans

        Returns:
            list -- list of plans associated with the region
        """
        return self._associated_plans

    def refresh(self):
        """Refresh the properties of the regions."""
        self._region_properties = self._get_region_properties()



# -*- coding: utf-8 -*-
# --------------------------------------------------------------------------
# Copyright Commvault Systems, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# --------------------------------------------------------------------------
""" File for operating on a MongoDB instance.
MongoDBInstance :   Derived class from BigDataAppsInstance Base class, representing a
                        MongoDBInstance instance and to perform operations on that instance
MongoDBInstance:
    __init__()                      --  Initializes MongoDB instance object with associated
    agent_object, instance name and instance id
    restore()                       -- Submits a restore request based on restore options
"""
from __future__ import unicode_literals
from ..bigdataappsinstance import BigDataAppsInstance
from ...exception import SDKException


class MongoDBInstance(BigDataAppsInstance):
    """
    Class for representing an Instance of the MongoDB  instance
    """
    def __init__(self, agent_object, instance_name, instance_id=None):
        """Initializes the object of the Mongo DB class
            Args:
                agent_object    (object)  --  instance of the Agent class
                instance_name   (str)     --  name of the instance
                instance_id     (str)     --  id of the instance
                    default: None
            Returns:
                object - instance of the Instance class
        """
        self._agent_object = agent_object
        self._browse_request = {}
        self._browse_url = None
        super(
            MongoDBInstance,
            self).__init__(
                agent_object,
                instance_name,
                instance_id)

    def restore(self, restore_options):
        """
            Restores the content of this instance content
            Args:
                restore_options : dict of keyword arguments needed to submit a MongoDB restore:
                    Example:
                       restore_options = {
                        restore_dict = {}
                            restore_dict["no_of_streams"] = 2
                            restore_dict["multinode_restore"] = True
                            restore_dict["destination_instance"] = self.client_name
                            restore_dict["destination_instance_id"] = self._instance_object.instance_id
                            restore_dict["paths"] = ["/"]
                            restore_dict["mongodb_restore"] = True
                            restore_dict["destination_client_id"] = self._client_obj.client_id
                            restore_dict["destination_client_name"] = self._client_obj.client_name
                            restore_dict["overwrite"] = True
                            restore_dict["client_type"] = 29
                            restore_dict["destination_appTypeId"] = 64
                            restore_dict["backupset_name"] = self.backupsetname
                            restore_dict["_type_"] = 5
                            restore_dict["subclient_id"] = -1
                            restore_dict["source_shard_name"] = self.replicaset
                            restore_dict["destination_shard_name"] = self.replicaset
                            restore_dict["hostname"] = self.primary_host
                            restore_dict["clientName"] = self.master_node
                            restore_dict["desthostName"] = self.primary_host
                            restore_dict["destclientName"] = self.master_node
                            restore_dict["destPortNumber"] = self.port
                            restore_dict["destDataDir"] = self.bin_path
                            restore_dict["bkpDataDir"] = self.bkp_dir_path
                            restore_dict["backupPortNumber"] = self.port
                            restore_dict["restoreDataDir"] = self.bkp_dir_path
                            restore_dict["primaryPort"] = self.port
                        }
            Returns:
                object - instance of the Job class for this restore job
        """
        if not (isinstance(restore_options, dict)):
            raise SDKException('Instance', '101')
        request_json = self._restore_json(restore_option=restore_options)

        request_json["taskInfo"]["associations"][0]["subclientId"] = restore_options.get(
            "subclient_id", -1)
        request_json["taskInfo"]["associations"][0]["backupsetName"] = restore_options.get(
            "backupset_name")
        request_json["taskInfo"]["associations"][0]["_type_"] = restore_options.get(
            "_type_")

        distributed_restore_json = {
            "distributedRestore": True,
        }


        client_object_source = self._commcell_object.clients.get(restore_options['clientName'])
        client_object_destination = self._commcell_object.clients.get(restore_options['destclientName'])
        distributed_restore_json["mongoDBRestoreOptions"] = {
                "destShardList": [
                    {
                        "srcShardName":  restore_options.get("source_shard_name", False),
                        "destShardName": restore_options.get("destination_shard_name", False),
                        "target": {
                            "hostName": restore_options.get("hostname", False),
                            "clientName": restore_options.get("clientName", False),
                            "clientId": int(client_object_source.client_id)
                        },
                        "destHostName": restore_options.get("desthostName", False),
                        "destPortNumber": restore_options.get("destPortNumber", False),
                        "destDataDir":  restore_options.get("restoreDataDir", False),
                        "bkpSecondary": {
                            "clientName": restore_options.get("clientName", False),
                            "hostName": restore_options.get("hostname", False),
                            "clientId": int(client_object_source.client_id)
                        },
                        "bkpHostName": restore_options.get("hostname", False),
                        "bkpPortNumber": restore_options.get("backupPortNumber", False),
                        "bkpDataDir": restore_options.get("bkpDataDir", False),
                        "useDestAsSecondary": False,
                        "primaryPortNumber":restore_options.get("primaryPort", False),
                    }
                ],
                "restoreFilesOnly": False,
                "recover": True,
                "pointInTimeToEndOfBackup": True,
                "latestOpLogSync": True,
                "latestEndOfBackup": True,
                "isGranularRecovery": False,
                "autoDBShutDown": True,
                "isInplaceRestore": True
            }

        request_json["taskInfo"]["subTasks"][0]["options"]["restoreOptions"][
                "distributedAppsRestoreOptions"] = distributed_restore_json
        return self._process_restore_response(request_json)



# -*- coding: utf-8 -*-

# --------------------------------------------------------------------------
# Copyright Commvault Systems, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# --------------------------------------------------------------------------

"""" Main file for performing the download operation

Download
========

    __init__(commcell_object)             --  initialize commcell_object of Download class
    associated with the commcell

    download_software()                   --  downloads software packages in the commcell

    sync_remote_cache()                   --  syncs remote cache

"""

from ..job import Job
from ..exception import SDKException
from .deploymentconstants import DownloadOptions
from cvpysdk.schedules import Schedules, SchedulePattern


class Download(object):
    """"class for downloading software packages"""

    def __init__(self, commcell_object):
        """Initialize commcell_object of the Download class.

            Args:
                commcell_object (object)  --  instance of the Commcell class

            Returns:
                object - instance of the Download class

        """

        self.commcell_object = commcell_object
        self._services = commcell_object._services
        self._cvpysdkcommcell_object = commcell_object._cvpysdk_object
        self.update_option = {}

    def download_software(
            self,
            options=None,
            os_list=None,
            service_pack=None,
            cu_number=0,
            sync_cache=True,
            sync_cache_list=None,
            schedule_pattern=None):
        """Downloads the os packages on the commcell

            Args:

                options      (enum)            --  Download option to download software

                os_list      (list of enum)    --  list of windows/unix packages to be downloaded

                service_pack (int)             --  service pack to be downloaded

                cu_number (int)                --  maintenance release number

                sync_cache (bool)              --  True if download and sync
                                                   False only download
                
                sync_cache_list (list)         --  list of names of remote caches to sync
                                                   use None to sync all caches

                schedule_pattern (dict)        --  pattern for schedule task
                                                   

            Returns:
                object - instance of the Job class for this download job

            Raises:
                SDKException:
                    if Download job failed

                    if response is empty

                    if response is not success

                    if another download job is running

            Usage:

            -   if download_software is not given any parameters it takes default value of latest
                service pack for options and downloads WINDOWS_64 package

                >>> commcell_obj.download_software()

            -   DownloadOptions and DownloadPackages enum is used for providing input to the
                download software method, it can be imported by

                >>> from cvpysdk.deployment.deploymentconstants import DownloadOptions
                    from cvpysdk.deployment.deploymentconstants import DownloadPackages

            -   sample method calls for different options, for latest service pack

                >>> commcell_obj.download_software(
                        options=DownloadOptions.lATEST_SERVICEPACK.value,
                        os_list=[DownloadPackages.WINDOWS_64.value]
                        )

            -   For Latest hotfixes for the installed service pack

                >>> commcell_obj.download_software(
                        options='DownloadOptions.LATEST_HOTFIXES.value',
                        os_list=[DownloadPackages.WINDOWS_64.value,
                                DownloadPackages.UNIX_LINUX64.value]
                        )

            -   For service pack and hotfixes

                >>> commcell_obj.download_software(
                        options='DownloadOptions.SERVICEPACK_AND_HOTFIXES.value',
                        os_list=[DownloadPackages.UNIX_MAC.value],
                        service_pack=13,
                        cu_number=42
                        )

                    **NOTE:** service_pack parameter must be specified for third option

        """

        # To set the default value if option is none
        version = self.commcell_object.commserv_version
        if options is None:
            options = 'latest service pack'

        if DownloadOptions.LATEST_SERVICEPACK.value == options:
            self.update_option = {
                'upgradeToLatestRelease': True,
                'latestFixesForCurrentRelease': False,
                "featureRelease": self.commcell_object.version,
                'SPName': 'latest',
                'IsSPName': False,
                'isSpDelayedDays': True,
                'isHotfixesDownload': False
            }
        elif DownloadOptions.LATEST_HOTFIXES.value == options:
            self.update_option = {
                'upgradeToLatestRelease': False,
                'latestFixesForCurrentRelease': True,
                "featureRelease": self.commcell_object.version,
                'SPName': 'hotfix',
                'IsSPName': False,
                'isSpDelayedDays': True,
                'isHotfixesDownload': True
            }
        elif DownloadOptions.SERVICEPACK_AND_HOTFIXES.value == options:
            if not (service_pack or cu_number):
                raise SDKException('Download', '102')
            self.update_option = {
                'upgradeToLatestRelease': False,
                'latestFixesForCurrentRelease': False,
                "featureRelease": "11." + service_pack + "." + cu_number
            }
        else:
            if service_pack is None:
                raise SDKException('Download', '102')

            self.update_option = {
                'SPName': str(service_pack),
                'IsSPName': True,
                'isSpDelayedDays': False,
                'isHotfixesDownload': False
            }

        # to set the default value if os_list is none
        if os_list is None:
            os_list = ['Windows(X64)']

        client_groups = []
        if sync_cache and sync_cache_list:
            for cache in sync_cache_list:
                client_groups.append({"type": "CLIENT_ENTITY",
                                      "clientName": cache,
                                      "id": self.commcell_object.clients.get(cache).client_id
                                      }) if version >= 36 else client_groups.append({"clientName": cache})
        elif sync_cache and not sync_cache_list:
            if version < 36:
                client_groups = [{"_type_": 2}]
            else:
                flag_1, response_1 = self._cvpysdkcommcell_object.make_request('GET', self._services['CREATE_RC'])
                cache_list = []
                for obj in response_1.json()['softwareCacheDetailList']:
                    cache_list.append(obj['cache']['name'])
                for client in cache_list:
                    client_groups.append({"type": "CLIENT_ENTITY",
                                          "clientName": client,
                                          "id": int(self.commcell_object.clients.all_clients[client]['id'])})

        if version >= 36:
            win_os_list = []
            unix_os_list = []
            temp_win = ['Windows(X64)', 'Windows(32)']

            for os in os_list:
                if os in temp_win:
                    win_os_list.append(os)
                else:
                    unix_os_list.append(os)
            request_json = {
                "downloadConfiguration": {
                    "upgradeToLatestRelease": self.update_option['upgradeToLatestRelease'],
                    "latestFixesForCurrentRelease": self.update_option['latestFixesForCurrentRelease'],
                    "featureRelease": self.update_option['featureRelease'],
                    "windowsDownloadOptions": win_os_list,
                    "unixDownloadOptions": unix_os_list,
                },
                "notifyWhenJobCompletes": False,
                "entities": client_groups

            }
        else:
            request_json = {
                "taskInfo": {
                    "task": {
                        "taskType": 1,
                        "initiatedFrom": 2,
                        "policyType": 0,
                        "alert": {
                            "alertName": ""
                        },
                        "taskFlags": {
                            "isEdgeDrive": False,
                            "disabled": False
                        }
                    },
                    "subTasks": [
                        {
                            "subTaskOperation": 1,
                            "subTask": {
                                "subTaskType": 1,
                                "operationType": 4019
                            },
                            "options": {
                                "adminOpts": {
                                    "updateOption": {
                                        "syncUpdateCaches": sync_cache,
                                        "spName": self.update_option['SPName'],
                                        "CUNumber": cu_number,
                                        "isWindows": True,
                                        "majorOnly": False,
                                        "isSpName": self.update_option['IsSPName'],
                                        "copyUpdates": True,
                                        "isHotfixesDownload": self.update_option['isHotfixesDownload'],
                                        "isSpDelayedDays": self.update_option['isSpDelayedDays'],
                                        "copySoftwareAndUpdates": False,
                                        "isUnix": True,
                                        "unixDownloadPackages": {
                                            "linuxosX64": 'Linux X86_64' in os_list,
                                            "solarisosX64": 'Solaris X86_64' in os_list,
                                            "solsparcos": 'Solaris-SPARC-X86' in os_list,
                                            "freeBSDos": 'Freebsd X86' in os_list,
                                            "linuxos": 'Linux X86' in os_list,
                                            "linuxosPPC64le": 'Linux PPC64le' in os_list,
                                            "freeBSDosX64": 'Freebsd X86_64' in os_list,
                                            "solarisos": 'Solaris SPARC' in os_list,
                                            "linuxs390os": 'Linux-S390-31' in os_list,
                                            "darwinos": 'macOS' in os_list,
                                            "linuxosS390": 'Linux-S390' in os_list,
                                            "aixppcos": 'Aix-PPC-32' in os_list,
                                            "linuxosPPC64": 'Linux-PPC-64' in os_list,
                                            "aixos": 'Aix PPC' in os_list,
                                            "hpos": 'HP IA64' in os_list,
                                            "solos": 'Solaris X86' in os_list
                                        },
                                        "windowsDownloadPackages": {
                                            "windowsX64": 'Windows(X64)' in os_list,
                                            "windows32": 'Windows(32)' in os_list
                                        },
                                        "clientAndClientGroups": client_groups,
                                        "downloadUpdatesJobOptions": {
                                            "downloadSoftware": True
                                        }
                                    }
                                }
                            }
                        }
                    ]
                }
            }

        if schedule_pattern:
            request_json = SchedulePattern().create_schedule(request_json, schedule_pattern)

        method = 'PUT' if version >= 36 else 'POST'
        url = self._services['DOWNLOAD_SOFTWARE'] if version >= 36 else self._services['CREATE_TASK']

        flag, response = self._cvpysdkcommcell_object.make_request(
            method, url, request_json
        )

        if flag:
            if response.json():
                if "jobId" in response.json() or "jobIds" in response.json():
                    return Job(self.commcell_object, response.json()['jobId']) if version >= 36 \
                        else Job(self.commcell_object, response.json()['jobIds'][0])

                elif "taskId" in response.json():
                    return Schedules(self.commcell_object).get(task_id=response.json()['taskId'])

                else:
                    raise SDKException('Download', '101')

            else:
                raise SDKException('Response', '102')
        else:
            raise SDKException('Response', '101')

    def copy_software(self, media_loc, username=None, password=None, sync_cache=True, schedule_pattern=None):
        """copies media from the specified location on the commcell

                    Args:

                        media_loc      (str)           --  Media Location to be used for copy software

                        username       (str)           --  username to authenticate to external location

                        password       (str)           --  password to authenticate to external location

                        sync_cache (bool)              --  True if download and sync
                                                           False only download

                        schedule_pattern(dict)         --  pattern for schedule task


                    Returns:
                        object - instance of the Job class for this copy software job

                    Raises:
                        SDKException:
                            if Download job failed

                            if response is empty

                            if response is not success

                            if another download job is running
                    Usage:

                        -   if media_location directory is local to the machine - username and password is not needed

                            >>> commcell_obj.copy_software(media_loc = "C:\\Downloads\\Media")

                        -   if Media_location directory is remote- username and passsword(base 64 encoded) are needed
                            to authenticate the cache

                            >>> commcell_obj.copy_software(
                            media_loc = "\\\\subdomain.company.com\\Media",
                            username = "domainone\\userone",
                            password = "base64encoded password"
                            )
                """
        client_auth = {}
        if username:
            if password is None:
                raise Exception(f"Password missing for remote location {media_loc}")
            client_auth = {
                "userName": username,
                "password": password
            }
        request_json = {
            "taskInfo": {
                "task": {
                    "taskType": 1,
                    "initiatedFrom": 2,
                    "policyType": 0,
                    "alert": {
                        "alertName": ""
                    },
                    "taskFlags": {
                        "isEdgeDrive": False,
                        "disabled": False
                    }
                },
                "subTasks": [
                    {
                        "subTaskOperation": 1,
                        "subTask": {
                            "subTaskType": 1,
                            "operationType": 4019
                        },
                        "options": {
                            "adminOpts": {
                                "updateOption": {
                                    "syncUpdateCaches": sync_cache,
                                    "copyUpdates": True,
                                    "copySoftwareAndUpdates": True,
                                    "clientAndClientGroups": [
                                        {
                                            "_type_": 2
                                        }
                                    ],
                                    "downloadUpdatesJobOptions": {
                                        "downloadSoftware": True,
                                        "updateCachePath": media_loc,
                                        "clientAuth": client_auth
                                    }
                                }
                            }
                        }
                    }
                ]
            }
        }

        if schedule_pattern:
            request_json = SchedulePattern().create_schedule(request_json, schedule_pattern)

        flag, response = self._cvpysdkcommcell_object.make_request(
            'POST', self._services['CREATE_TASK'], request_json
        )

        if flag:
            if response.json():
                if "jobIds" in response.json():
                    return Job(self.commcell_object, response.json()['jobIds'][0])

                elif "taskId" in response.json():
                    return Schedules(self.commcell_object).get(task_id=response.json()['taskId'])

                else:
                    raise SDKException('Download', '101')

            else:
                raise SDKException('Response', '102')
        else:
            raise SDKException('Response', '101')

    def sync_remote_cache(self, client_list=None):
        """Syncs remote cache

            Args:

                client_list  --  list of client names
                Default is None. By default all remote cache clients are synced

            Returns:
                object - instance of the Job class for sync job

            Raises:
                SDKException:
                    if sync job failed

                    if response is empty

                    if response is not success

                    if another sync job is running with the given client

        """

        clients = []
        if client_list is None:

            clients = [{
                "_type_": 2
            }]
        else:
            for each in client_list:
                clients.append({
                    "_type_": 3,
                    "clientName": each})

        request_json = {
            "taskInfo": {
                "task": {
                    "ownerId": 1,
                    "taskType": 1,
                    "ownerName": "admin",
                    "initiatedFrom": 1,
                    "policyType": 0,
                    "taskFlags": {
                        "disabled": False
                    }
                },
                "subTasks": [
                    {
                        "subTaskOperation": 1,
                        "subTask": {
                            "subTaskType": 1,
                            "operationType": 4019
                        },
                        "options": {
                            "adminOpts": {
                                "updateOption": {
                                    "syncUpdateCaches": True,
                                    "invokeLevel": 1,
                                    "isWindows": False,
                                    "majorOnly": False,
                                    "isSpName": False,
                                    "copyUpdates": True,
                                    "isHotfixesDownload": False,
                                    "isSpDelayedDays": True,
                                    "copySoftwareAndUpdates": False,
                                    "isUnix": False,
                                    "clientAndClientGroups": clients,
                                    "downloadUpdatesJobOptions": {
                                        "downloadSoftware": False
                                    }
                                }
                            }
                        }
                    }
                ]
            }
        }

        flag, response = self._cvpysdkcommcell_object.make_request(
            'POST', self._services['CREATE_TASK'], request_json
        )

        if flag:
            if response.json():
                if "jobIds" in response.json():
                    return Job(self.commcell_object, response.json()['jobIds'][0])

                else:
                    raise SDKException('Download', '101')

            else:
                raise SDKException('Response', '102')
        else:
            raise SDKException('Response', '101')



# -*- coding: utf-8 -*-

# --------------------------------------------------------------------------
# Copyright Commvault Systems, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# --------------------------------------------------------------------------

"""File for performing tags related operations.


Tags: Class for representing all the tags created by the user which is logged in.


Tags:
    Methods:

        has_tag()   --  checks whether the tag with given name exists or not

        add()       --  creates a new tag to commcell

        get()       --  returns a tag object for given tag name

        delete()    --  deletes an entity tag from commcell

        refresh()   --  refreshes the tags list

    Properties:

        **all_tags  --  returns the dict containing all the tags and id

Tag:
    Properties:

        **tag_id    --  returns tag id

        **tag_name  --  returns tag name

"""
from cvpysdk.exception import SDKException


class Tags:
    """Class for doing operations related to entity tags from backend"""

    DEFAULT_TAGSET_ID = -1

    def __init__(self, commcell_object: object) -> None:
        """Method to initialize tags class

            Args:
                commcell_object (Commcell)  -- instance of commcell class
            Returns:
                Tags (object)               -- instance of tags class
        """
        self._commcell_object = commcell_object
        self._service = commcell_object._services
        self._cvpysdk_object = commcell_object._cvpysdk_object
        self._get_tags_url = self._service['GET_ENTITY_TAGS']
        self._create_tags_url = self._service['CREATE_ENTITY_TAGS']
        self._tags = None

        self.refresh()

    def _get_tags(self) -> dict:
        """Gets all the tags created by a user

            Returns:
                tags: name-id pair of tags associated to a user
                    Example:
                        {
                            "tag1": 1,
                            "tag2": 3,
                            "tag3": 4
                        }
            Raises:
                SDKException:
                    if response is empty

                    if response is not success
        """
        flag, response = self._cvpysdk_object.make_request('GET', self._get_tags_url)

        if not flag:
            response_string = self._commcell_object._update_response_(response.text)
            raise SDKException('Response', '101', response_string)

        resp = response.json()
        if not resp:
            raise SDKException('Response', '102')

        self.DEFAULT_TAGSET_ID = resp['tagSetInfo']['id']

        tag_dict = {}
        if 'tags' in resp:
            for tag in resp.get("tags", []):
                tag_dict[tag.get("name").lower()] = tag.get("id")
            return tag_dict

    @property
    def all_tags(self) -> dict:
        """Returns a dictionary containing tags and their ID associated to a user

            Returns:
                tags: name-id pair of tags associated to a user
                    Example:
                        {
                            "tag1": 1,
                            "tag2": 3,
                            "tag3": 4
                        }
        """

        return self._tags

    def has_tag(self, tag_name: str) -> bool:
        """Checks if a tag exist for the logged-in user

            Args:
                tag_name (str): name of the tag to search

            Returns:
                bool: return True if tag is found, else returns false

            Raises:
                SDKException:
                    if type of the tag name argument is not string

        """
        if not isinstance(tag_name, str):
            raise SDKException('EntityTags', '101')

        return self._tags and tag_name.lower() in self._tags

    def get(self, name: str):
        """Returns an instance of the Tag class for the given tag name.

            Args:
                name    (str)   --  name of the tag to get the instance of

            Returns:
                object  -   instance of the Tag class for the given tag name

            Raises:
                SDKException:
                    if type of the tag name argument is not string

                    if no tag exists with the given name

        """
        if self.has_tag(name):
            name = name.lower()
            return Tag(self._commcell_object, name, self._tags[name])

        raise SDKException('EntityTags', '105')

    def refresh(self) -> None:
        """Refresh the list of tags"""
        self._tags = self._get_tags()

    def add(self, tag_name: str):
        """Method to add an entity tag

            Args:
                tag_name (str): entity tag name

            Returns:
                object  -   instance of the Tag class, for the newly created entity tag
        """
        if not isinstance(tag_name, str):
            raise SDKException('EntityTags', '101')

        payload = {
            'container': {
                'containerId': self.DEFAULT_TAGSET_ID
                },
            'tags': [
                {
                    'name': f"{tag_name}"
                }
            ]
        }

        flag, response = self._cvpysdk_object.make_request('POST', self._service['CREATE_ENTITY_TAGS'], payload)

        if not flag:
            response_string = self._commcell_object._update_response_(response.text)
            raise SDKException('Response', '101', response_string)

        response_json = response.json()
        if 'errList' in response_json:
            err_message = response_json.get('errLogMessage')
            err_code = response_json.get('errorCode')

            if err_code or err_message:
                raise SDKException('EntityTags', '103', err_message)

        self.refresh()
        return self.get(tag_name)

    def delete(self, tag_name: str) -> None:
        """Deletes the entity tag

        Args:
            tag_name (str): entity tag name to delete
        """
        if not self.has_tag(tag_name):
            raise SDKException("EntityTags", '105')

        tag_name = tag_name.lower()
        tag_id = self._tags[tag_name]

        flag, response = self._cvpysdk_object.make_request('DELETE', self._service['DELETE_ENTITY_TAGS'] % tag_id)

        if not flag:
            response_string = self._commcell_object._update_response_(response.text)
            raise SDKException('Response', '101', response_string)

        resp = response.json()
        err_code = 0
        err_message = ""
        if resp:
            err_code = resp.get('errorCode', 0)
            err_message = resp.get('errorMessage', "")

        if err_code or err_message:
            raise SDKException("EntityTags", '102', f"Error: {err_message}")

        self.refresh()


class Tag:
    """Class for performing actions related to tag """

    def __init__(self, commcell_object, tag_name, tag_id=None):
        """Initialise the Tag class instance.

            Args:
                commcell_object     (object)    --  instance of the Commcell class

                tag_name            (str)       --  name of the entity tag

                tag_id              (str)       --  id of the entity tag
                    default: None

            Returns:
                object  -   instance of the tag class

        """
        self._commcell_object = commcell_object

        self._cvpysdk_object = commcell_object._cvpysdk_object
        self._services = commcell_object._services
        self._update_response_ = commcell_object._update_response_

        self._tag_name = tag_name

        if tag_id:
            self._tag_id = str(tag_id)
        else:
            self._tag_id = self._get_tag_id()

    def _get_tag_id(self) -> str:
        """Gets tag id based on the tag name"""
        tags = Tags(self._commcell_object)
        return tags.get(self._tag_name).tag_id

    @property
    def tag_id(self):
        """Returns tag id"""
        return self._tag_id

    @property
    def tag_name(self):
        """Returns tag name"""
        return self._tag_name
